
"""Test dockerhub authentication"""
import pytest

from mldock.platform_helpers.docker import auth
from mldock.platform_helpers.utils import set_env

from tests.mocks import docker as mock_docker
from tests.mocks.aws import boto3 as mock_boto3
from tests.mocks.google.auth import default
from tests.mocks.google.auth import Request

class TestDockerAuth:

    def init_mocks(self, mocker):


        mocker.patch(
            'mldock.platform_helpers.docker.auth.docker',
            mock_docker
        )


        mocker.patch(
            'mldock.platform_helpers.aws.auth.boto3',
            mock_boto3
        )

        mocker.patch(
            'mldock.platform_helpers.gcp.auth.default',
            default
        )

        mocker.patch(
            'mldock.platform_helpers.gcp.auth.Request',
            Request
        )

    def testdocker_auth_dockerhub_success(self, mocker):

        self.init_mocks(mocker)

        env_vars = {
            "DOCKERHUB_USERNAME": "USERNAME",
            "DOCKERHUB_PASSWORD": "PASSWORD",
            "DOCKERHUB_REGISTRY": "https://index.docker.io/v1/",
            "DOCKERHUB_REPO": "REPONAME"
        }
        valid_vars = [{"key": key, 'value': value} for key,value in env_vars.items()]

        with set_env(**env_vars):

            (
                client,
                metadata
            ) = auth.login_and_authenticate(
                provider='dockerhub', region=None
            )

            assert metadata['username'] == 'USERNAME'

            assert metadata['password'] == 'PASSWORD'

            assert metadata['registry'] == 'https://index.docker.io/v1/'

            assert metadata['repository'] == 'REPONAME'

    def test_docker_auth_gcr_success(self, mocker):

        self.init_mocks(mocker)

        (
            client,
            metadata
        ) = auth.login_and_authenticate(
            provider='gcr', region='eu'
        )

        assert metadata['username'] ==  'oauth2accesstoken'

        assert metadata['password'] == 'PASSWORD'

        assert metadata['registry'] == 'https://eu.gcr.io/myproject'

        assert metadata['repository'] == 'eu.gcr.io/myproject'

    def test_docker_auth_ecr_success(self, mocker):

        self.init_mocks(mocker)

        (
            client,
            metadata
        ) = auth.login_and_authenticate(
            provider='ecr', region='eu'
        )

        assert metadata['username'] ==  'USERNAME'

        assert metadata['password'] == 'PASSWORD'

        assert metadata['registry'] == 'ABCXYZ.dkr.ecr.us-west-2.amazonaws.com'

        assert metadata['repository'] == 'ABCXYZ.dkr.ecr.us-west-2.amazonaws.com'

    def test_docker_auth_not_supported_platform_fail(self, mocker):

        self.init_mocks(mocker)
        provider = 'UNKNOWN'
        with pytest.raises(
            TypeError,
            match=r"{PROVIDER} Registry is not implemented.".format(PROVIDER=provider)
        ) as excinfo:
            (
                client,
                metadata
            ) = auth.login_and_authenticate(
                provider=provider, region='eu'
            )



import logging
from pathlib import Path
import pandas as pd
from sklearn import datasets

from mldock.platform_helpers.mldock.configuration.environment.gcp import GCPEnvironment
from mldock.platform_helpers.mldock.configuration.container import \
    BaseTrainingContainer, BaseServingContainer

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('mldock')

# Instantiate Environment here
# import to other scripts
environment = GCPEnvironment(base_dir='/opt/ml/')

# (TODO)
# Assess situations and likelihood of whether users should by
# default create their own or can it be hidden until needed.
# For example, using just the BaseServingContainer by default
# USE-CASES:
#   - special Metric loggers (cleaning up, uploading to separate s3 locatations, etc)
#   - Environment and importing from here to other scripts
class TrainingContainer(BaseTrainingContainer):
    """Implements the base training container, allow a user to override/add/extend any training container setup logic"""
    
    def startup(self):

        super(TrainingContainer, self).startup()
        # in 'dev' we can add the following
        if self.environment.environment_variables('MLDOCK_STAGE', default=None) == "dev":

            # download data
            iris = datasets.load_iris(as_frame=True)
            X = iris.data
            y = iris.target
            # write data to iris input data channel
            pd.concat([X, y], axis=1).to_csv(Path(self.environment.input_data_dir,'iris/train.csv'), index=False)

# (TODO)
# Assess situations and likelihood of whether users should by
# default create their own or can it be hidden until needed.
# For example, using just the BaseServingContainer by default
# USE-CASES:
#   - special Metric loggers (cleaning up, uploading to separate s3 locatations, etc)
#   - Instantiating Environment and importing from here to other scripts
class ServingContainer(BaseServingContainer):
    """Implements the base serving container, allow a user to override/add/extend any serving container setup logic"""
    pass



from tests.mocks.google.auth import default
from tests.mocks.google.auth import Request

from mldock.platform_helpers.gcp import auth

class TestGCPDockerAuth:

    def init_mocks(self, mocker):

        mocker.patch(
            'mldock.platform_helpers.gcp.auth.default',
            default
        )

        mocker.patch(
            'mldock.platform_helpers.gcp.auth.Request',
            Request
        )

    def test_get_gcp_gcr_success(self, mocker):

        self.init_mocks(mocker)

        username, password, registry, repository = auth.get_gcp_gcr(region='eu')

        assert username == 'oauth2accesstoken'

        assert password == 'PASSWORD'

        assert registry == 'https://eu.gcr.io/myproject'

        assert repository == 'eu.gcr.io/myproject'



import sys
import traceback

def extract_stack_trace() -> str:
    """Extracts full stacktrace from thown exception
    Returns:
        str: traceback of error thrown
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()
    stack_trace = traceback.format_exception(
        exc_type,
        exc_value,
        exc_traceback
    )
    return repr(stack_trace)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
#
# Disclaimer: This code can be found here: https://github.com/aws/sagemaker-training-toolkit/blob/master/test/unit/test_encoder.py
#
import io
import itertools

from mock import Mock, patch
import numpy as np
import pytest
from scipy import sparse
from six import BytesIO

from mldock.platform_helpers.mldock.inference import content_types
from mldock.platform_helpers.mldock.inference.content_encoders import numpy as numpy_encoders
from mldock.platform_helpers.mldock.inference.record_pb2 import Record
from mldock.platform_helpers.mldock.inference.recordio import _read_recordio

class TestNumpyEncoders:
    """Tests the numpy encoder methods"""

    @staticmethod
    @pytest.mark.parametrize(
        "target",
        [
            [42, 6, 9],
            [42.0, 6.0, 9.0],
            ["42", "6", "9"],
            [u"42", u"6", u"9"],
            {42: {"6": 9.0}}
        ],
    )
    def test_array_to_npy(target):
        """test numpy arrays are correctly encoded as .npy"""
        input_data = np.array(target)

        actual = numpy_encoders.array_to_npy(input_data)

        np.testing.assert_equal(np.load(BytesIO(actual), allow_pickle=True), np.array(target))

        actual = numpy_encoders.array_to_npy(target)

        np.testing.assert_equal(np.load(BytesIO(actual), allow_pickle=True), np.array(target))

    @staticmethod
    @pytest.mark.parametrize(
        "target, expected",
        [
            ([42, 6, 9], "[42, 6, 9]"),
            ([42.0, 6.0, 9.0], "[42.0, 6.0, 9.0]"),
            (["42", "6", "9"], '["42", "6", "9"]'),
            ({42: {"6": 9.0}}, '{"42": {"6": 9.0}}'),
        ],
    )
    def test_array_to_json(target, expected):
        """test numpy arrays are correctly encoded as json"""
        actual = numpy_encoders.array_to_json(target)
        np.testing.assert_equal(actual, expected)

        actual = numpy_encoders.array_to_json(np.array(target))
        np.testing.assert_equal(actual, expected)

    @staticmethod
    @pytest.mark.parametrize(
        "target, expected, quoted",
        [
            ([42.0, 6.0, 9.0], '42.0\n6.0\n9.0\n', False),
            (["42", "6", "9"], '42\n6\n9\n', False),
            ([42, 6, 9], '"42"\n"6"\n"9"\n', True),
            ([42.0, 6.0, 9.0], '"42.0"\n"6.0"\n"9.0"\n', True),
            (["42", "6", "9"], '"42"\n"6"\n"9"\n', True),
            (["False,", "True.", "False,"], '"False,"\n"True."\n"False,"\n', True),
            (["aaa", 'b"bb', "ccc"], '"aaa"\n"b"bb"\n"ccc"\n', True),
            (["a\nb", "c"], '"a\nb"\n"c"\n', True),
        ],
    )
    def test_array_to_csv(target, expected, quoted):
        """test numpy arrays are correctly encoded as csv"""
        actual = numpy_encoders.array_to_csv(target, quoted=quoted)
        np.testing.assert_equal(actual, expected)

        actual = numpy_encoders.array_to_csv(np.array(target).astype(str), quoted=quoted)
        np.testing.assert_equal(actual, expected)

    @staticmethod
    @pytest.mark.parametrize(
        "array_data",
        [
            [[1.0, 2.0, 3.0], [10.0, 20.0, 30.0]]
        ],
    )
    def test_array_to_recordio_dense(array_data):
        """test numpy arrays are correctly encoded as dense recordio"""
        buf = numpy_encoders.array_to_recordio_protobuf(np.array(array_data))
        stream = io.BytesIO(buf)

        for record_data, expected in zip(_read_recordio(stream), array_data):
            record = Record()
            record.ParseFromString(record_data)
            assert record.features["values"].float64_tensor.values == expected

    @staticmethod
    @pytest.mark.parametrize(
        "array_data, keys_data",
        [
            (
                [[1.0, 2.0], [10.0, 30.0], [100.0, 200.0, 300.0, 400.0], [1000.0, 2000.0, 3000.0]], [[0, 1], [1, 2], [0, 1, 2, 3], [0, 2, 3]]
            )
        ],
    )
    def test_sparse_int_write_spmatrix_to_sparse_tensor(array_data, keys_data):
        """test numpy arrays are correctly encoded to int sparse tensor spmatrix encoding"""
        n = 4

        flatten_data = list(itertools.chain.from_iterable(array_data))
        y_indices = list(itertools.chain.from_iterable(keys_data))
        x_indices = [[i] * len(keys_data[i]) for i in range(len(keys_data))]
        x_indices = list(itertools.chain.from_iterable(x_indices))

        array = sparse.coo_matrix((flatten_data, (x_indices, y_indices)), dtype="int")
        buf = numpy_encoders.array_to_recordio_protobuf(array)
        stream = io.BytesIO(buf)

        for record_data, expected_data, expected_keys in zip(
            _read_recordio(stream), array_data, keys_data
        ):
            record = Record()
            record.ParseFromString(record_data)
            assert record.features["values"].int32_tensor.values == expected_data
            assert record.features["values"].int32_tensor.keys == expected_keys
            assert record.features["values"].int32_tensor.shape == [n]

    @staticmethod
    @pytest.mark.parametrize(
        "array_data, keys_data",
        [
            (
                [[1.0, 2.0], [10.0, 30.0], [100.0, 200.0, 300.0, 400.0], [1000.0, 2000.0, 3000.0]], [[0, 1], [1, 2], [0, 1, 2, 3], [0, 2, 3]]
            )
        ],
    )
    def test_sparse_float32_write_spmatrix_to_sparse_tensor(array_data, keys_data):
        """test numpy arrays are correctly encoded to float32 sparse tensor spmatrix"""
        n = 4

        flatten_data = list(itertools.chain.from_iterable(array_data))
        y_indices = list(itertools.chain.from_iterable(keys_data))
        x_indices = [[i] * len(keys_data[i]) for i in range(len(keys_data))]
        x_indices = list(itertools.chain.from_iterable(x_indices))

        array = sparse.coo_matrix((flatten_data, (x_indices, y_indices)), dtype="float32")
        buf = numpy_encoders.array_to_recordio_protobuf(array)
        stream = io.BytesIO(buf)

        for record_data, expected_data, expected_keys in zip(
            _read_recordio(stream), array_data, keys_data
        ):
            record = Record()
            record.ParseFromString(record_data)
            assert record.features["values"].float32_tensor.values == expected_data
            assert record.features["values"].float32_tensor.keys == expected_keys
            assert record.features["values"].float32_tensor.shape == [n]



import tests.mocks.aws.boto3 as mock_boto3

from mldock.platform_helpers.aws import auth

class TestAWSDockerAuth:

    def init_mocks(self, mocker):


        mocker.patch(
            'mldock.platform_helpers.aws.auth.boto3',
            mock_boto3
        )

    def test_get_aws_ecr(self, mocker):

        self.init_mocks(mocker)

        username, password, registry, repository = auth.get_aws_ecr(region='eu-west-1')

        assert username == 'USERNAME'

        assert password == 'PASSWORD'

        assert registry == 'ABCXYZ.dkr.ecr.us-west-2.amazonaws.com'

        assert repository == 'ABCXYZ.dkr.ecr.us-west-2.amazonaws.com'



"""
    Project Container Config manager

    Config Managers that create and update the container project config.
    Currently supported:
        - mldock
"""
import os
import json
import click
import logging
from pathlib import Path

from mldock.terminal import ChoiceWithNumbers, style_dropdown
from mldock.config_managers.core import BaseConfigManager

logger=logging.getLogger('mldock')

class MLDockConfigManager(BaseConfigManager):
    """Hyperparameter Config Manager for mldock
    """

    config = {}

    def setup_config(self, **kwargs):
        click.secho("Base Setup", bg='blue')
        self.ask_for_image_name()
        self.ask_for_template_name()
        self.ask_for_mldock_module_dir()
        self.ask_for_container_dir_name()
        self.ask_for_requirements_file_name()
    
    def update_config(self, **kwargs):
        self.config.update(kwargs)

    def ask_for_image_name(self):
        """prompt user for image name
        """
        image_name = click.prompt(
            text=click.style("Set your image name: ", fg='bright_blue'),
            default=self.config.get('image_name', 'my_ml_container')
        )

        self.config.update({
            'image_name':image_name
        })

    def ask_for_template_name(self):
        """prompt user for container template name
        """

        click.secho("Set container template name", bg='blue', nl=True)
        options = ['generic', 'gcp', 'aws']

        template_name = click.prompt(
            text=style_dropdown(
                group_name="template name",
                options=options,
                default=self.config.get('template', 'generic')
            ),
            type=ChoiceWithNumbers(options, case_sensitive=False),
            show_default=True,
            default=self.config.get('template', 'generic'),
            show_choices=False
        )
        
        self.config.update({
            'template': template_name
        })

    def ask_for_container_dir_name(self):
        """prompt user for container dir name
        """
        container_dir_name = click.prompt(
            text=click.style("Set your container dir name: ", fg='bright_blue'),
            default=self.config.get('container_dir', 'container')
        )

        self.config.update({
            'container_dir': container_dir_name
        })

    def ask_for_mldock_module_dir(self):
        """prompt user for mldock module dir
        """
        
        mldock_module_dir = click.prompt(
            text=click.style("Set mldock module dir: ", fg='bright_blue'),
            default=self.config.get('mldock_module_dir', 'src')
        )

        self.config.update({
            'mldock_module_dir': mldock_module_dir
        })
    
    @staticmethod
    def _format_data_node_item(item, base_path='data'):

        new_key = item["channel"]
        new_value = os.path.join(base_path, item["channel"], item["filename"])
        return "\t{KEY} : {VALUE}".format(KEY=new_key, VALUE=new_value)

    def _format_data_node(self):

        output = []
        
        for item in self.config['data']:
            output.append(self._format_data_node_item(item, base_path='data'))

        return "\n".join(output)

    def _format_model_node(self):

        output = []
        
        for item in self.config['model']:
            output.append(self._format_data_node_item(item, base_path='model'))

        return "\n".join(output)

    def _format_stage_node(self):

        output = []
        
        for key_, value_ in self.config['stages'].items():
            image_and_tag = "{}:{}".format(self.config['image_name'], value_["tag"])
            output.append("\t{} : {}".format(key_, image_and_tag))

        return "\n".join(output)

    @staticmethod
    def _format_nodes(config):

        output = []
        
        for key_, value_ in config.items():
            output.append("\t{} : {}".format(key_, value_))

        return "\n".join(output)

    def get_state(self):
        """pretty prints a json config to terminal
        """
        config = self.config.copy()

        config.pop("stages")
        formatted_stages = self._format_stage_node()
        hyperparameters = config.pop("hyperparameters")
        config.pop("data")
        formatted_data_node = self._format_data_node()
        config.pop("model")
        formatted_model_node = self._format_model_node()
        environment = config.pop("environment")

        states = []

        states.append({"name": "Base Setup", "message": self._format_nodes(config)})
        states.append({"name": "Data Channels", "message": formatted_data_node})
        states.append({"name": "Model Channels", "message": formatted_model_node})
        states.append({"name": "Stages", "message": formatted_stages})
        states.append({"name": "Hyperparameters", "message": self._format_nodes(hyperparameters)})
        states.append({"name": "Environment Variables", "message": self._format_nodes(environment)})

        return states

    def ask_for_requirements_file_name(self):
        """prompt user for image name
        """
        requirements_file_name = click.prompt(
            text=click.style("Set full path to requirements: ", fg='bright_blue'),
            default=self.config.get('requirements_dir', 'requirements.txt')
        )

        self.config.update({
            'requirements_dir':requirements_file_name
        })
    
    def update_stages(self, stages: dict):
        """"""
        self.config.update(
            {'stages': stages}
        )

    def update_env_vars(self, environment: dict):
        """
            Update environment node in .mldock config

            Args:
                environment (dict): key,value dictionaries
        """
        self.config.update(
            {'environment': environment}
        )

    def update_hyperparameters(self, hyperparameters: dict):
        """
            Update hyperparameter node in .mldock config

            Args:
                hyperparameters (dict): key,value dictionaries
        """
        self.config.update(
            {'hyperparameters': hyperparameters}
        )

    def update_data_channels(self, data: dict):
        """
            Update data node in .mldock config

            Args:
                data (dict): key,value dictionaries
        """
        self.config.update(
            {'data': data}
        )
    
    def update_model_channels(self, models: dict):
        """
            Update data node in .mldock config

            Args:
                data (dict): key,value dictionaries
        """
        self.config.update(
            {'model': models}
        )



from mldock.platform_helpers.mldock.utils import \
    _format_dictionary_as_env_vars, _format_key_as_mldock_env_var, \
        collect_mldock_environment_variables

class TestUtils:

    sample_environment = {
        "run_id": 2,
        "A_DUMMY_CONDITIONAL": True
    }

    sample_hyperparameters = {
        'hyperparameters': {
            "factors": 10,
            "Alpha criterion": 'gini',
            "weighting": {'A': 10, 'C': 2}
        }
    }

    sample_hyperparameter_validation = {'MLDOCK_HYPERPARAMETERS': '{"factors": 10, "Alpha criterion": "gini", "weighting": {"A": 10, "C": 2}}'}

    sample_collected_mldock_env_vars = {'MLDOCK_ENVIRONMENT': '{"run_id": 2, "A_DUMMY_CONDITIONAL": true}', 'MLDOCK_HYPERPARAMETERS': '{"factors": 10, "Alpha criterion": "gini", "weighting": {"A": 10, "C": 2}}', 'MLDOCK_STAGE': 'dev'}
    def test_format_key_as_mldock_env_var(self):
        """test format key for environment vars"""

        result = _format_key_as_mldock_env_var(
            key='run_id',
            prefix='mldock'
        )

        assert result == 'MLDOCK_RUN_ID', (
            "Failed, Keys were not in MLDOCK format"
        )


    def test_format_dictionary_as_env_vars(self):
        """test format key for environment vars"""

        result = _format_dictionary_as_env_vars(
            obj=self.sample_hyperparameters,
            group='mldock'
        )

        assert result == self.sample_hyperparameter_validation, (
            "Failed, Keys were not in MLDOCK format"
        )

    def test_collect_mldock_environment_variables(self):
        """test format key for environment vars"""

        result = collect_mldock_environment_variables(
            stage='dev',
            hyperparameters=self.sample_hyperparameters['hyperparameters'],
            environment=self.sample_environment
        )

        assert result == self.sample_collected_mldock_env_vars, (
            "Failed, Keys were not in MLDOCK format"
        )



""" 
    Environment Handling Script
    (Only needed for training environment)
        - handles loading vars from ENV vars in container
        - handles create directories for data channels
        - handles getting container specs e.g. num_gpus, num_cpus, etc

        This module contains functionality related to the container environment.
        This includes:
                - constants
                - utility functions
                - classes which provide access to relevant aspects of the environment (including system characteristics,
                        filesystem locations, environment variables, and configuration files).

"""
from __future__ import absolute_import

import json
import logging
import multiprocessing
import os
import shlex
import socket
import subprocess
import sys
import time
import environs

from sagemaker_training import logging_config, mapping, params
from mldock.platform_helpers import utils

logger = logging_config.get_logger()

SAGEMAKER_BASE_PATH = os.path.join("/opt", "ml")
BASE_PATH_ENV = "SAGEMAKER_BASE_DIR"


def _write_json(obj, path):  # type: (object, str) -> None
    """Write a serializeable object as a JSON file."""
    with open(path, "w") as f:
        json.dump(obj, f)


def _is_training_path_configured():  # type: () -> bool
    """Check if the tree structure with data and configuration files used for training
    exists.
    When a SageMaker Training Job is created, the Docker container that will be used for
    training is executed with the folder /opt/ml attached. The /opt/ml folder contains
    data and configurations files necessary for training.
    Outside SageMaker, the environment variable SAGEMAKER_BASE_DIR defines the location
    of the base folder.
    This function checks whether /opt/ml exists or if the base folder variable exists.
    Returns:
        (bool): Whether the training path is configured.
    """
    return os.path.exists(SAGEMAKER_BASE_PATH) or BASE_PATH_ENV in os.environ


def _set_base_path_env():  # type: () -> None
    """Set the environment variable SAGEMAKER_BASE_DIR as
    ~/sagemaker_local/{timestamp}/opt/ml
    """
    local_config_dir = os.path.join(
        os.path.expanduser("~"), "sagemaker_local", "jobs", str(time.time()), "opt", "ml"
    )

    logger.info("Setting environment variable SAGEMAKER_BASE_DIR as %s ." % local_config_dir)
    os.environ[BASE_PATH_ENV] = local_config_dir


_is_path_configured = _is_training_path_configured()

if not _is_path_configured:
    logger.info("Directory /opt/ml does not exist.")
    _set_base_path_env()

base_dir = os.environ.get(BASE_PATH_ENV, SAGEMAKER_BASE_PATH)  # type: str

code_dir = os.path.join(base_dir, "code")
"""str: the path of the user's code directory, e.g., /opt/ml/code/"""

model_dir = os.path.join(base_dir, "model")  # type: str
"""str: the directory where models should be saved, e.g., /opt/ml/model/"""

input_dir = os.path.join(base_dir, "input")  # type: str
"""str: the path of the input directory, e.g. /opt/ml/input/
The input_dir, e.g. /opt/ml/input/, is the directory where SageMaker saves input data
        and configuration files before and during training.
        The input data directory has the following subdirectories:
            config (`input_config_dir`) and data (`_input_data_dir`)
        Returns:
            str: the path of the input directory, e.g. /opt/ml/input/
"""

_input_data_dir = os.path.join(input_dir, "data")  # type: str

input_config_dir = os.path.join(input_dir, "config")  # type: str
"""str: the path of the input directory, e.g. /opt/ml/input/config/
The directory where standard SageMaker configuration files are located, e.g. /opt/ml/input/config/.
SageMaker training creates the following files in this folder when training starts:
    - `hyperparameters.json`: Amazon SageMaker makes the hyperparameters in a CreateTrainingJob
        request available in this file.
    - `inputdataconfig.json`: You specify data channel information in the InputDataConfig parameter
        in a CreateTrainingJob request. Amazon SageMaker makes this information available
        in this file.
    - `resourceconfig.json`: name of the current host and all host containers in the training
More information about this files can be find here:
    https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html
Returns:
    str: the path of the input directory, e.g. /opt/ml/input/config/
"""

output_dir = os.path.join(base_dir, "output")  # type: str
"""str: the path to the output directory, e.g. /opt/ml/output/.
The directory where training success/failure indications will be written, e.g. /opt/ml/output.
To save non-model artifacts check `output_data_dir`.
Returns:
    str: the path to the output directory, e.g. /opt/ml/output/.
"""

output_data_dir = os.path.join(output_dir, "data")  # type: str

output_intermediate_dir = os.path.join(output_dir, "intermediate")  # type: str
"""str: the path to the intermediate output directory, e.g. /opt/ml/output/intermediate.
The directory special behavior is to move artifacts from the training instance to
s3 directory during training.
Returns:
    str: the path to the intermediate output directory, e.g. /opt/ml/output/intermediate.
"""

HYPERPARAMETERS_FILE = "hyperparameters.json"  # type: str
RESOURCE_CONFIG_FILE = "resourceconfig.json"  # type: str
INPUT_DATA_CONFIG_FILE = "inputdataconfig.json"  # type: str

hyperparameters_file_dir = os.path.join(input_config_dir, HYPERPARAMETERS_FILE)  # type: str
input_data_config_file_dir = os.path.join(input_config_dir, INPUT_DATA_CONFIG_FILE)  # type: str
resource_config_file_dir = os.path.join(input_config_dir, RESOURCE_CONFIG_FILE)  # type: str

""" REUSABLE UTILITIES
"""

def _create_training_directories():
    """Create the directory structure and files necessary for training under the base path.
    """
    logger.info("Creating a new training folder under %s ." % base_dir)

    os.makedirs(model_dir)
    os.makedirs(input_config_dir)
    os.makedirs(output_data_dir)

    _write_json({}, hyperparameters_file_dir)
    _write_json({}, input_data_config_file_dir)

    host_name = socket.gethostname()

    resources_dict = {"current_host": host_name, "hosts": [host_name]}
    _write_json(resources_dict, resource_config_file_dir)


if not _is_path_configured:
    _create_training_directories()


def _create_code_dir():  # type: () -> None
    """Create /opt/ml/code when the module is imported."""
    if not os.path.exists(code_dir):
        os.makedirs(code_dir)


_create_code_dir()

def read_hyperparameters():  # type: () -> dict
    """Read the hyperparameters from /opt/ml/input/config/hyperparameters.json.
    For more information about hyperparameters.json:
    https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html#your-algorithms-training-algo-running-container-hyperparameters
    Returns:
         (dict[string, object]): A dictionary containing the hyperparameters.
    """
    hyperparameters = utils._read_json(hyperparameters_file_dir)

    deserialized_hps = {}

    for k, v in hyperparameters.items():
        try:
            v = json.loads(v)
        except (ValueError, TypeError):
            logger.info(
                "Failed to parse hyperparameter %s value %s to Json.\n"
                "Returning the value itself",
                k,
                v,
            )

        deserialized_hps[k] = v

    return deserialized_hps


def read_resource_config():  # type: () -> dict
    """Read the resource configuration from /opt/ml/input/config/resourceconfig.json.
    For more information about resourceconfig.json:
https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html#your-algorithms-training-algo-running-container-dist-training
    Returns:
        resource_config (dict[string, object]): the contents from /opt/ml/input/config/resourceconfig.json.
                        It has the following keys:
                            - current_host: The name of the current container on the container
                                            network. For example, 'algo-1'.
                            -  hosts: The list of names of all containers on the container
                                      network, sorted lexicographically. For example,
                                      `['algo-1', 'algo-2', 'algo-3']` for a three-node cluster.
    """
    return utils._read_json(resource_config_file_dir)


def read_input_data_config():  # type: () -> dict
    """Read the input data configuration from /opt/ml/input/config/inputdataconfig.json.
        For example, suppose that you specify three data channels (train, evaluation, and
        validation) in your request. This dictionary will contain:
        {'train': {
            'ContentType':  'trainingContentType',
            'TrainingInputMode': 'File',
            'S3DistributionType': 'FullyReplicated',
            'RecordWrapperType': 'None'
        },
        'evaluation' : {
            'ContentType': 'evalContentType',
            'TrainingInputMode': 'File',
            'S3DistributionType': 'FullyReplicated',
            'RecordWrapperType': 'None'
        },
        'validation': {
            'TrainingInputMode': 'File',
            'S3DistributionType': 'FullyReplicated',
            'RecordWrapperType': 'None'
        }}
        For more information about inpudataconfig.json:
https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html#your-algorithms-training-algo-running-container-dist-training
    Returns:
            input_data_config (dict[string, object]): Contents from /opt/ml/input/config/inputdataconfig.json.
    """
    return utils._read_json(input_data_config_file_dir)


def channel_path(channel):  # type: (str) -> str
    """ Return the directory containing the channel data file(s) which is:
    - <self.base_dir>/input/data/<channel>
    For more information about channels: https://docs.aws.amazon.com/sagemaker/latest/dg/API_Channel.html
    Returns:
        str: The input data directory for the specified channel.
    """
    return os.path.join(_input_data_dir, channel)


def num_gpus():  # type: () -> int
    """Return the number of GPUs available in the current container.
    Returns:
        int: Number of GPUs available in the current container.
    """
    try:
        cmd = shlex.split("nvidia-smi --list-gpus")
        output = subprocess.check_output(cmd).decode("utf-8")
        return sum([1 for x in output.split("\n") if x.startswith("GPU ")])
    except (OSError, subprocess.CalledProcessError):
        logger.info("No GPUs detected (normal if no gpus installed)")
        return 0


def num_cpus():  # type: () -> int
    """Return the number of CPUs available in the current container.
    Returns:
        int: Number of CPUs available in the current container.
    """
    return multiprocessing.cpu_count()


class Environment(mapping.MappingMixin):  # pylint:disable=too-many-public-methods
    """Provides access to aspects of the training environment relevant to training jobs, including
    hyperparameters, system characteristics, filesystem locations, environment variables and
    configuration settings.
    The Environment is a read-only snapshot of the container environment during training. It does
    not contain any form of state.
    It is a dictionary like object, allowing any builtin function that works with dictionary.
    Example on how a script can use training environment:
            >>>from sagemaker_training import environment
            >>>env = environment.Environment()
            get the path of the channel 'training' from the inputdataconfig.json file
            >>>training_dir = environment.channel_input_dirs['training']
            get a the hyperparameter 'training_data_file' from hyperparameters.json file
            >>>file_name = environment.hyperparameters['training_data_file']
            get the folder where the model should be saved
            >>>model_dir = environment.model_dir
            >>>data = np.load(os.path.join(training_dir, file_name))
            >>>x_train, y_train = data['features'], keras.utils.to_categorical(data['labels'])
            >>>model = ResNet50(weights='imagenet')
            ...
            >>>model.fit(x_train, y_train)
            save the model in the end of training
            >>>model.save(os.path.join(model_dir, 'saved_model'))
    Attributes:
        current_host (str): The name of the current container on the container network. For
                            example, 'algo-1'.
        module_name (str): The name of the user provided module.
        module_dir (str): The full path location of the user provided module.
        input_dir (str): The input_dir, e.g. /opt/ml/input/, is the directory where SageMaker saves
                         input data and configuration files before and during training. The input
                         data directory has the following subdirectories:
                         config (`input_config_dir`) and data (`_input_data_dir`)
        input_config_dir (str): The directory where standard SageMaker configuration files are
                                located, e.g. /opt/ml/input/config/.
            SageMaker training creates the following files in this folder when training starts:
                - `hyperparameters.json`: Amazon SageMaker makes the hyperparameters in a
                                          CreateTrainingJob request available in this file.
                - `inputdataconfig.json`: You specify data channel information in the
                                          InputDataConfig parameter in a CreateTrainingJob request.
                                          Amazon SageMaker makes this information available in this
                                          file.
                - `resourceconfig.json`: name of the current host and all host containers in the
                                         training
            More information about these files can be find here:
                https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html
        model_dir (str):  the directory where models should be saved, e.g., /opt/ml/model/
        output_dir (str): The directory where training success/failure indications will be written,
                          e.g. /opt/ml/output. To save non-model artifacts check `output_data_dir`.
        hyperparameters (dict[string, object]): An instance of `HyperParameters` containing the
                                                training job hyperparameters.
        resource_config (dict[string, object]): the contents from
                                                /opt/ml/input/config/resourceconfig.json.
            It has the following keys:
                - current_host: The name of the current container on the container network.
                    For example, 'algo-1'.
                -  hosts: The list of names of all containers on the container network,
                    sorted lexicographically. For example, `['algo-1', 'algo-2', 'algo-3']`
                    for a three-node cluster.
        input_data_config (dict[string, object]): the contents from /opt/ml/input/config/inputdataconfig.json.
            For example, suppose that you specify three data channels (train, evaluation, and
            validation) in your request. This dictionary will contain:
            {'train': {
                'ContentType':  'trainingContentType',
                'TrainingInputMode': 'File',
                'S3DistributionType': 'FullyReplicated',
                'RecordWrapperType': 'None'
            },
            'evaluation' : {
                'ContentType': 'evalContentType',
                'TrainingInputMode': 'File',
                'S3DistributionType': 'FullyReplicated',
                'RecordWrapperType': 'None'
            },
            'validation': {
                'TrainingInputMode': 'File',
                'S3DistributionType': 'FullyReplicated',
                'RecordWrapperType': 'None'
            }}
            You can find more information about /opt/ml/input/config/inputdataconfig.json here:
            https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html#your-algorithms-training-algo-running-container-inputdataconfig
        output_data_dir (str): The dir to write non-model training artifacts (e.g. evaluation
                               results) which will be retained by SageMaker,
                               e.g. /opt/ml/output/data. As your algorithm runs in a container,
                               it generates output including the status of the training job and
                               model and output artifacts. Your algorithm should write this
                               information to the this directory.
        hosts (list[str]): The list of names of all containers on the container network, sorted
                           lexicographically. For example, `['algo-1', 'algo-2', 'algo-3']` for
                           a three-node cluster.
        channel_input_dirs (dict[string, string]): containing the data channels and the directories
                                                   where the training data was saved. When you run
                                                   training, you can partition your training data
                                                   into different logical 'channels'. Depending on
                                                   your problem, some common channel ideas are:
                                                   'train', 'test', 'evaluation' or 'images',
                                                   'labels'.
            The format of channel_input_dir is as follows:
                - `channel`(str) - the name of the channel defined in the input_data_config.
                - `training data path`(str) - the path to the directory where the training data is
                saved.
        framework_module (str):  Name of the framework module and entry point. For example:
            my_module:main
        network_interface_name (str): Name of the network interface used for distributed training.
        job_name (str): The name of the current training job.
    """

    def __init__(self, resource_config=None, input_data_config=None, hyperparameters=None):
        """Initialize a read-only snapshot of the container environment.
        Args:
            resource_config (dict[string, object]): The contents from
                /opt/ml/input/config/resourceconfig.json.
                It has the following keys:
                    - current_host: The name of the current container on the container network.
                        For example, 'algo-1'.
                    -  hosts: The list of names of all containers on the container network,
                        sorted lexicographically. For example, `['algo-1', 'algo-2', 'algo-3']`
                        for a three-node cluster.
            input_data_config (dict[string, object]): The contents from /opt/ml/input/config/inputdataconfig.json.
                For example, suppose that you specify three data channels (train, evaluation, and
                validation) in your request. This dictionary will contain:
                {'train': {
                    'ContentType':  'trainingContentType',
                    'TrainingInputMode': 'File',
                    'S3DistributionType': 'FullyReplicated',
                    'RecordWrapperType': 'None'
                },
                'evaluation' : {
                    'ContentType': 'evalContentType',
                    'TrainingInputMode': 'File',
                    'S3DistributionType': 'FullyReplicated',
                    'RecordWrapperType': 'None'
                },
                'validation': {
                    'TrainingInputMode': 'File',
                    'S3DistributionType': 'FullyReplicated',
                    'RecordWrapperType': 'None'
                }}
                You can find more information about /opt/ml/input/config/inputdataconfig.json here:
                https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html#your-algorithms-training-algo-running-container-inputdataconfig
            hyperparameters (dict[string, object]): An instance of `HyperParameters` containing the
                training job hyperparameters.
        """
        current_host = os.environ.get(params.CURRENT_HOST_ENV)
        module_name = os.environ.get(params.USER_PROGRAM_ENV, None)
        module_dir = os.environ.get(params.SUBMIT_DIR_ENV, code_dir)
        log_level = int(os.environ.get(params.LOG_LEVEL_ENV, logging.INFO))
        self.environment_variables = environs.Env()
        self.environment_variables.read_env()

        self._current_host = current_host
        self._num_gpus = num_gpus()
        self._num_cpus = num_cpus()
        self._module_name = module_name
        self._user_entry_point = module_name
        self._module_dir = module_dir
        self._log_level = log_level
        self._model_dir = model_dir

        resource_config = resource_config or read_resource_config()
        input_data_config = input_data_config or read_input_data_config()

        self.setup_hyperparameters()
        all_hyperparameters = hyperparameters or read_hyperparameters()

        current_host = resource_config["current_host"]
        hosts = resource_config["hosts"]

        split_result = mapping.split_by_criteria(
            all_hyperparameters,
            keys=params.SAGEMAKER_HYPERPARAMETERS,
            prefix=params.SAGEMAKER_PREFIX,
        )
        sagemaker_hyperparameters = split_result.included

        additional_framework_parameters = {
            k: sagemaker_hyperparameters[k]
            for k in sagemaker_hyperparameters.keys()
            if k not in params.SAGEMAKER_HYPERPARAMETERS
        }

        sagemaker_region = sagemaker_hyperparameters.get(
            params.REGION_NAME_PARAM, "sagemaker_region"
        )

        os.environ[params.JOB_NAME_ENV] = sagemaker_hyperparameters.get(params.JOB_NAME_PARAM, "")
        os.environ[params.CURRENT_HOST_ENV] = current_host
        os.environ[params.REGION_NAME_ENV] = sagemaker_region or ""

        self._hosts = hosts

        # eth0 is the default network interface defined by SageMaker with VPC support and
        # local mode.
        # ethwe is the current network interface defined by SageMaker training, it will be
        # changed to eth0 in the short future.
        self._network_interface_name = resource_config.get("network_interface_name", "eth0")

        self._hyperparameters = split_result.excluded
        self._additional_framework_parameters = additional_framework_parameters
        self._resource_config = resource_config
        self._input_data_config = input_data_config
        self._output_data_dir = output_data_dir
        self._output_intermediate_dir = output_intermediate_dir
        self._channel_input_dirs = {channel: channel_path(channel) for channel in input_data_config}
        self._current_host = current_host

        # override base class attributes
        if self._module_name is None:
            self._module_name = str(sagemaker_hyperparameters.get(params.USER_PROGRAM_PARAM, None))
        self._user_entry_point = self._user_entry_point or sagemaker_hyperparameters.get(
            params.USER_PROGRAM_PARAM
        )

        self._module_dir = str(sagemaker_hyperparameters.get(params.SUBMIT_DIR_PARAM, code_dir))
        self._log_level = sagemaker_hyperparameters.get(params.LOG_LEVEL_PARAM, logging.INFO)
        self._sagemaker_s3_output = sagemaker_hyperparameters.get(
            params.S3_OUTPUT_LOCATION_PARAM, None
        )
        self._framework_module = os.environ.get(params.FRAMEWORK_TRAINING_MODULE_ENV, None)

        self._input_dir = input_dir
        self._input_config_dir = input_config_dir
        self._output_dir = output_dir
        self._job_name = os.environ.get(params.TRAINING_JOB_ENV.upper(), None)

        self._master_hostname = list(hosts)[0]
        self._is_master = current_host == self._master_hostname

    def setup_hyperparameters(self):
        """Retrieves the env vars matching hyperparameters regex and updates config"""
        hyperparameters = utils._read_json(
            hyperparameters_file_dir
        )

        updated_hparams = self.environment_variables.json('MLDOCK_HYPERPARAMETERS', default='{}')

        hyperparameters.update(
            updated_hparams
        )

        utils._write_json(
            obj=hyperparameters,
            path=hyperparameters_file_dir
        )

    @property
    def model_dir(self):  # type: () -> str
        """ The directory where models should be saved.
        Returns:
            str: The directory where models should be saved, e.g., /opt/ml/model/
        """
        return self._model_dir

    @property
    def current_host(self):  # type: () -> str
        """The name of the current container on the container network. For example, 'algo-1'.
        Returns:
            str: Current host.
        """
        return self._current_host

    @property
    def num_gpus(self):  # type: () -> int
        """The number of GPUs available in the current container.
        Returns:
            int: Number of GPUs available in the current container.
        """
        return self._num_gpus

    @property
    def num_cpus(self):  # type: () -> int
        """The number of CPUs available in the current container.
        Returns:
            int: Number of CPUs available in the current container.
        """
        return self._num_cpus

    @property
    def module_name(self):  # type: () -> str
        """The name of the user provided module.
        Returns:
            str: Name of the user provided module.
        """
        return self._parse_module_name(self._module_name)

    @property
    def module_dir(self):  # type: () -> str
        """The full path location of the user provided module.
        Returns:
            str: Full path location of the user provided module.
        """
        return self._module_dir

    @property
    def log_level(self):  # type: () -> int
        """Environment logging level.
        Returns:
            int: Environment logging level.
        """
        return self._log_level

    @property
    def user_entry_point(self):  # type: () -> str
        """The name of provided user entry point.
        Returns:
            str: The name of provided user entry point.
        """
        return self._user_entry_point

    @staticmethod
    def _parse_module_name(program_param):
        """Given a module name or a script name, Returns the module name.
        This function is used for backwards compatibility.
        Args:
            program_param (str): Module or script name.
        Returns:
            str: Module name.
        """
        if program_param and program_param.endswith(".py"):
            return program_param[:-3]
        return program_param

    @property
    def is_master(self):  # type: () -> bool
        """Returns True if host is master.
        """
        return self._is_master

    @property
    def master_hostname(self):  # type: () -> str
        """Returns the hostname of the master node.
        """
        return self._master_hostname

    @property
    def job_name(self):  # type: () -> str
        """The name of the current training job.
        Returns:
            str: The training job name.
        """
        return self._job_name

    @property
    def additional_framework_parameters(self):  # type: () -> dict
        """The dict of additional framework hyperparameters. All the hyperparameters prefixed with
        'sagemaker_' but not in SAGEMAKER_HYPERPARAMETERS will be included here.
        Returns:
            dict: Additional framework hyperparameters, SageMaker Python SDK adds hyperparameters
                  with a prefix **sagemaker_** during training. These hyperparameters are
                  framework independent settings and are not defined by the user.
        """
        return self._additional_framework_parameters

    def sagemaker_s3_output(self):  # type: () -> str
        """S3 output directory location provided by the user.
        Returns:
            str: S3 location.
        """
        return self._sagemaker_s3_output

    def to_cmd_args(self):
        """Command line arguments representation of the training environment.
        Returns:
            (list): List of cmd arguments.
        """
        return mapping.to_cmd_args(self.hyperparameters)

    def to_env_vars(self):
        """Environment variable representation of the training environment.
        Returns:
            dict: An instance of dictionary.
        """

        env = {
            "hosts": self.hosts,
            "network_interface_name": self.network_interface_name,
            "hps": self.hyperparameters,
            "user_entry_point": self.user_entry_point,
            "framework_params": self.additional_framework_parameters,
            "resource_config": self.resource_config,
            "input_data_config": self.input_data_config,
            "output_data_dir": self.output_data_dir,
            "channels": sorted(self.channel_input_dirs.keys()),
            "current_host": self.current_host,
            "module_name": self.module_name,
            "log_level": self.log_level,
            "framework_module": self.framework_module,
            "input_dir": self.input_dir,
            "input_config_dir": self.input_config_dir,
            "output_dir": self.output_dir,
            "num_cpus": self.num_cpus,
            "num_gpus": self.num_gpus,
            "model_dir": self.model_dir,
            "module_dir": self.module_dir,
            "training_env": dict(self),
            "user_args": self.to_cmd_args(),
            "output_intermediate_dir": self.output_intermediate_dir,
        }

        for name, path in self.channel_input_dirs.items():
            env["channel_%s" % name] = path

        for key, value in self.hyperparameters.items():
            env["hp_%s" % key] = value

        return mapping.to_env_vars(env)

    @property
    def hosts(self):  # type: () -> list
        """The list of names of all containers on the container network, sorted lexicographically.
                For example, `["algo-1", "algo-2", "algo-3"]` for a three-node cluster.
        Returns:
              list[str]: All the hosts in the training network.
        """
        return self._hosts

    @property
    def channel_input_dirs(self):  # type: () -> dict
        """A dict[str, str] containing the data channels and the directories where the training
        data was saved.
        When you run training, you can partition your training data into different logical
        "channels".
        Depending on your problem, some common channel ideas are: "train", "test", "evaluation"
            or "images',"labels".
        The format of channel_input_dir is as follows:
            - `channel`[key](str) - the name of the channel defined in the input_data_config.
            - `training data path`[value](str) - the path to the directory where the training
                                                 data is saved.
        Returns:
            dict[str, str] With the information about the channels.
        """
        return self._channel_input_dirs

    @property
    def network_interface_name(self):  # type: () -> str
        """Name of the network interface used for distributed training.
        Returns:
              str: Name of the network interface, for example, 'ethwe'.
        """
        return self._network_interface_name

    @property
    def input_dir(self):  # type: () -> str
        """The input_dir, e.g. /opt/ml/input/, is the directory where SageMaker saves input data
        and configuration files before and during training.
        The input data directory has the following subdirectories:
            config (`input_config_dir`) and data (`input_data_dir`)
        Returns:
            str: The path of the input directory, e.g. /opt/ml/input/.
        """
        return self._input_dir

    @property
    def input_config_dir(self):  # type: () -> str
        """The directory where standard SageMaker configuration files are located, e.g.
        /opt/ml/input/config/.
        SageMaker training creates the following files in this folder when training starts:
            - `hyperparameters.json`: Amazon SageMaker makes the hyperparameters in a
                                      CreateTrainingJob request available in this file.
            - `inputdataconfig.json`: You specify data channel information in the
                                      InputDataConfig parameter in a CreateTrainingJob request.
                                      Amazon SageMaker makes this information available in this
                                      file.
            - `resourceconfig.json`: name of the current host and all host containers in the
                                     training More information about this files can be find here:
            https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html
        Returns:
            str: The path of the input directory, e.g. /opt/ml/input/config/.
        """
        return self._input_config_dir

    @property
    def output_dir(self):  # type: () -> str
        """The directory where training success/failure indications will be written,
        e.g. /opt/ml/output.
        To save non-model artifacts check `output_data_dir`.
        Returns:
            str: The path to the output directory, e.g. /opt/ml/output/.
        """
        return self._output_dir

    @property
    def hyperparameters(self):  # type: () -> dict
        """The dict of hyperparameters that were passed to the training job.
        Returns:
            dict[str, object]: An instance of `HyperParameters` containing the training job
                                                    hyperparameters.
        """
        return self._hyperparameters

    @property
    def resource_config(self):  # type: () -> dict
        """A dictionary with the contents from /opt/ml/input/config/resourceconfig.json.
        It has the following keys:
            - current_host: The name of the current container on the container
                            network. For example, 'algo-1'.
            -  hosts: The list of names of all containers on the container network,
                      sorted lexicographically. For example,
                      `["algo-1", "algo-2", "algo-3"]` for a three-node cluster.
        Returns:
            dict[str, str or list(str)]
        """
        return self._resource_config

    @property
    def input_data_config(self):  # type: () -> dict
        """A dictionary with the contents from /opt/ml/input/config/inputdataconfig.json.
        For example, suppose that you specify three data channels (train,
        evaluation, and validation) in your request. This dictionary will contain:
        ```{"train": {
                "ContentType":  "trainingContentType",
                "TrainingInputMode": "File",
                "S3DistributionType": "FullyReplicated",
                "RecordWrapperType": "None"
            },
            "evaluation" : {
                "ContentType": "evalContentType",
                "TrainingInputMode": "File",
                "S3DistributionType": "FullyReplicated",
                "RecordWrapperType": "None"
            },
            "validation": {
                "TrainingInputMode": "File",
                "S3DistributionType": "FullyReplicated",
                "RecordWrapperType": "None"
            }
         } ```
        You can find more information about /opt/ml/input/config/inputdataconfig.json here:
            https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html#your-algorithms-training-algo-running-container-inputdataconfig
        Returns:
            dict[str, dict[str, str]]
        """
        return self._input_data_config

    @property
    def output_data_dir(self):  # type: () -> str
        """The dir to write non-model training artifacts (e.g. evaluation results) which will be
        retained by SageMaker, e.g. /opt/ml/output/data/{current_host}.
        As your algorithm runs in a container, it generates output including the status of the
        training job and model and output artifacts. Your algorithm should write this information
        to the this directory.
        Returns:
            str: The path to output data directory, e.g. /opt/ml/output/data/algo-1.
        """
        return self._output_data_dir

    @property
    def output_intermediate_dir(self):  # type: () -> str
        """The directory for intermediate output artifacts that should be synced to S3.
        Any files written to this directory will be uploaded to S3 by a background process
        while training is in progress, but only if sagemaker_s3_output was specified.
        Returns:
            str: The path to the intermediate output directory, e.g. /opt/ml/output/intermediate.
        """
        return self._output_intermediate_dir

    @property
    def framework_module(self):  # type: () -> str
        """Name of the framework module and entry point.
        Returns:
            str: Name of the framework module and entry point. For example:
                my_module:main
        """
        return self._framework_module


def write_env_vars(env_vars=None):  # type: (dict) -> None
    """Write the dictionary env_vars in the system, as environment variables.
    Args:
        env_vars (dict): A dictionary of environment variables.
    """
    env_vars = env_vars or {}
    env_vars["PYTHONPATH"] = ":".join(sys.path)

    for name, value in env_vars.items():
        os.environ[name] = value



"""Test AWS s3 storage helpers"""
from pathlib import Path
import tempfile
from mock import patch
from mldock.platform_helpers.aws import storage
from tests.mocks.aws.boto3.s3 import AWSBucket, Objects, S3Resource, AWSBucketObject

class TestAWSStorage:
    """Test aws s3 storage helpers"""

    @staticmethod
    def setup_storage_mocks(bucket_name, prefix_path, filename, storage_destination, mock_boto3):
        """
            setup up storage mocks for S3 storage

            returns:
                (bucket: AWSBucket, blob: AWSBucketObject)
        """
        aws_bucket_objects = Objects(bucket_name=bucket_name, files=[filename])
        bucket = AWSBucket(name=bucket_name, objects=aws_bucket_objects)
        bucket_object = AWSBucketObject(
            key=Path(prefix_path, filename).as_posix(),
            bucket_name=bucket_name
        )
        s3_resource = S3Resource(resource='s3', bucket=bucket)

        mock_boto3.resource.return_value = s3_resource

        return bucket, bucket_object

    @patch('mldock.platform_helpers.aws.storage.boto3')
    @patch('mldock.platform_helpers.aws.storage.download_file_from_bucket')
    def test_download_file_from_bucket_recieved_correct_args(self, mock_download_file_from_bucket, mock_boto3):
        """
            test that download_file_from_bucket recieves the correct arguments
            during download_folder.
        """
        bucket_name = 'some-bucket'
        prefix_path = 'some/prefix'
        filename = 'file.txt'
        local_path = 'tmp/local/path'
        storage_destination = Path(prefix_path, 'file.txt').as_posix()

        bucket, bucket_object = self.setup_storage_mocks(
            bucket_name, prefix_path, filename, storage_destination, mock_boto3
        )

        storage.download_folder(
            bucket_name=bucket_name,
            prefix=prefix_path,
            target='tmp/local/path'
        )

        args, kwargs = mock_download_file_from_bucket.call_args
        print(args, kwargs)
        # maybe get args given?
        assert args ==(bucket, bucket_object, 'some/prefix', 'file.txt', 'tmp/local/path'), (
            "Failed. download_file_from_bucket recieved the wrong arguments"
        )

    @patch('mldock.platform_helpers.aws.storage._iter_nested_dir')
    @patch('mldock.platform_helpers.aws.storage.boto3')
    @patch('mldock.platform_helpers.aws.storage.upload_file_to_bucket')
    def test_upload_file_to_bucket_recieved_correct_args(self, mock_upload_file_to_bucket, mock_boto3, mock__iter_nested_dir):
        """
            test that upload_file_to_bucket recieves the correct arguments
            during upload_folder.
        """
        bucket_name = 'some-bucket'
        prefix_path = 'some/prefix'
        filename = 'file.txt'
        storage_destination = Path(prefix_path, 'file.txt').as_posix()

        with tempfile.TemporaryDirectory() as tmp_dir:
            local_path = tmp_dir
            path = Path(local_path, 'file.txt')
            path.touch()

            bucket, bucket_object = self.setup_storage_mocks(
                bucket_name, prefix_path, filename, storage_destination, mock_boto3
            )

            mock__iter_nested_dir.return_value = [path]

            storage.upload_folder(
                local_path=local_path,
                bucket_name=bucket_name,
                prefix=prefix_path
            )
            args, kwargs = mock_upload_file_to_bucket.call_args

        # maybe get args given?
        assert args == (bucket, path, storage_destination), (
            "Failed. upload_file_to_bucket recieved the wrong arguments"
        )



"""Local API methods"""
import os
import sys
import json
import time
import logging
from pathlib import Path
import requests
import docker
import traceback
import logging


from mldock.config_managers.cli import \
    CliConfigureManager
from mldock.platform_helpers import utils

logger=logging.getLogger('mldock')

# log formatting
def pretty_build_logs(line: dict):

    if line is None:
        return None

    error = line.get('error', None)
    errorDetail = line.get('errorDetail', None)

    if error is not None:
        logger.error('{}\n{}'.format(error, errorDetail))
    
    stream = line.get('stream', '')

    if ('Step' in stream) & (':' in stream):
        logger.info(" ==> {MESSAGE}".format(MESSAGE=stream))
    else:
        logger.debug(" ==> {MESSAGE}".format(MESSAGE=stream))


class DockerLLCManager:

    def __init__(self, **kwargs):
        self.kwargs = kwargs 
    def __enter__(self):
        self.client = docker.APIClient(**self.kwargs)
        return self.client
    def __exit__(self, exc_type, exc_value, exc_tb):
        logger.info(self.client.prune_builds())
        self.client.close()

class DockerManager:

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __enter__(self):
        self.client = docker.from_env(**self.kwargs)
        return self.client
    def __exit__(self, exc_type, exc_value, exc_tb):
        self.client.close()

# API methods

def train_model(working_dir, docker_tag, image_name, entrypoint, cmd, env=None):
    """
    Trains ML model(s) locally
    :param working_dir: [str], source root directory
    :param docker_tag: [str], the Docker tag for the image
    :param image_name: [str], The name of the Docker image
    """
    process_env = {}
    if env is not None:
        process_env.update(env)

    config_manager = CliConfigureManager()

    try:
        with DockerManager() as client:
            base_ml_path = '/opt/ml'

            container_volumes = {
                Path(working_dir,'config').absolute().as_posix(): {'bind': Path(base_ml_path,'input/config').as_posix(), 'mode': 'rw'},
                Path(working_dir,'data').absolute().as_posix(): {'bind': Path(base_ml_path,'input/data').as_posix(), 'mode': 'rw'},
                Path(working_dir,'model').absolute().as_posix(): {'bind': Path(base_ml_path,'model').as_posix(), 'mode': 'rw'},
                Path(working_dir,'output').absolute().as_posix(): {'bind': Path(base_ml_path,'output').as_posix(), 'mode': 'rw'}
            }

            if config_manager.local.get('auth_type', None) is not None:

                credentials_volume = utils.get_sdk_credentials_volume_mount(config_manager.local.get('auth_type'))
                container_volumes.update(credentials_volume)

            container = client.containers.run(
                image="{IMAGE}:{TAG}".format(IMAGE=image_name, TAG=docker_tag),
                entrypoint=entrypoint,
                command=cmd,
                environment=process_env,
                remove=True,
                tty=True,
                volumes=container_volumes,
                detach=True,
                stream=True
            )
            logs = container.logs(follow=True).decode('utf-8')

            logger.info(logs)
    except (KeyboardInterrupt, SystemExit) as exception:
        logger.error(exception)
        container.kill()
        raise
    except (docker.errors.APIError, docker.errors.ContainerError, docker.errors.ImageNotFound) as exception:
        logger.error(exception)
        raise
    except Exception as exception:
        logger.error(exception)
        raise

def deploy_model(working_dir, docker_tag, image_name, entrypoint, cmd, port=8080, env={}, **kwargs):
    """
    Deploys ML models(s) locally
    :param working_dir: [str], source root directory
    :param docker_tag: [str], the Docker tag for the image
    :param image_name: [str], The name of the Docker image
    """
    verbose = kwargs.get('verbose', False)
    process_env = {}
    if isinstance(env, dict) and len(env) > 0:
        process_env.update(env)

    config_manager = CliConfigureManager()
    try:
        with DockerManager() as client:
            base_ml_path = '/opt/ml'
            container_volumes = {
                Path(working_dir,'config').absolute().as_posix(): {'bind': Path(base_ml_path,'input/config').as_posix(), 'mode': 'rw'},
                Path(working_dir,'data').absolute().as_posix(): {'bind': Path(base_ml_path,'input/data').as_posix(), 'mode': 'rw'},
                Path(working_dir,'model').absolute().as_posix(): {'bind': Path(base_ml_path,'model').as_posix(), 'mode': 'rw'},
                Path(working_dir,'output').absolute().as_posix(): {'bind': Path(base_ml_path,'output').as_posix(), 'mode': 'rw'}
            }

            if config_manager.local.get('auth_type', None) is not None:

                credentials_volume = utils.get_sdk_credentials_volume_mount(config_manager.local.get('auth_type'))
                container_volumes.update(credentials_volume)

            container = client.containers.run(
                image="{IMAGE}:{TAG}".format(IMAGE=image_name, TAG=docker_tag),
                entrypoint=entrypoint,
                command=cmd,
                environment=process_env,
                ports={8080: port},
                remove=True,
                tty=True,
                volumes=container_volumes,
                auto_remove=True,
                detach=True,
                stream=True
            )
            if verbose:
                logs = container.logs(follow=True).decode('utf-8')

                logger.info(logs)

    except requests.exceptions.ConnectionError as exception:
        logger.info(
            "\nWorker socket connection timed out. "
            "\n\nThis can happen when connection pool is overloaded. "
            "We suggest setting environment variable: DOCKER_CLIENT_TIMEOUT=120 to give the client more time."
        )

    except (KeyboardInterrupt, SystemExit) as exception:
        logger.error(exception)
        container.kill()
        raise
    except (docker.errors.APIError, docker.errors.ContainerError, docker.errors.ImageNotFound) as exception:
        logger.error(exception)
        raise
    except Exception as exception:
        logger.error(exception)
        raise

def docker_build(
    image_name: str,
    dockerfile_path: str,
    module_path: str,
    target_dir_name: str,
    requirements_file_path: str,
    no_cache: bool,
    docker_tag: str,
    container_platform: str = None
):
    """Runs the build executable using docker python sdk.

    Args:
        script_path (str): relative path to script when run on root
        base_path (str):
        image_name (str):
        dockerfile_path (str):
        module_path (str):
        target_dir_name (str):
        requirements_file_path (str):
        docker_tag (str): the docker tag to build
    """
    try:

        with DockerLLCManager(
            max_pool_size=20,
            base_url=os.environ.get('DOCKER_HOST','unix:///var/run/docker.sock')
        ) as client:

            logs = client.build(
                tag="{IMAGE}:{TAG}".format(IMAGE=image_name, TAG=docker_tag),
                path='.',
                dockerfile=Path(os.path.join(dockerfile_path, 'Dockerfile')).as_posix(),
                buildargs={
                    'module_path': Path(module_path).as_posix(),
                    'target_dir_name': Path(target_dir_name).as_posix(),
                    'requirements_file_path': Path(requirements_file_path).as_posix(),
                    'container_platform': container_platform
                },
                nocache=no_cache,
                rm=True,
                decode=True
            )
            for line in logs:
                yield line
            #     pretty_build_logs(line=line)

    except (KeyboardInterrupt, SystemExit) as exception:
        logger.error(exception)
        raise
    except (docker.errors.DockerException) as exception:
        if "NewConnectionError" in str(exception):
            raise Exception("Unable to connect to docker daemon. Set environment variable: DOCKER_HOST")
        raise
    except (docker.errors.APIError, docker.errors.ContainerError, docker.errors.ImageNotFound) as exception:
        logger.error(exception)
        raise
    except Exception as exception:
        logger.error(exception)
        raise

def search_mldock_containers(**kwargs):
    """Performs a docker ps with filters. Incorporating the label to find only MLDOCK Containers"""
    mldock_filters = {}
    client = docker.from_env()
    ancestor = kwargs.get("ancestor", None)
    if ancestor is not None:
        mldock_filters.update({
            "ancestor": ancestor
        })
    
    label = kwargs.get("label", ["MLDOCK__IS_MLDOCK_CONTAINER=true"])
    mldock_filters.update(
        {
            "label": label
        }
    )
    containers = client.containers.list(filters=mldock_filters)

    return containers



import os
import io
import csv
import logging
import numpy as np
from mldock.platform_helpers.mldock.errors import extract_stack_trace
from mldock.platform_helpers.mldock.model_service import base
from src.container.assets import environment, logger

class ModelService(base.ModelService):
    model = None

    def load_model(self, model_path):
        """Get the model object for this instance, loading it if it's not already loaded."""
        if self.model == None:
            # TODO place your model load code here
            # get your model from MODEL_DIR
            self.model = None
        return self.model

    @staticmethod
    def input_transform(input):
        """
            Custom input transformer

            args:
                input: raw input from request
        """
        return input

    @staticmethod
    def output_transform(predictions):
        """
            Custom output transformation code
        """
        return predictions

    def predict(self, input):
        """For the input, do the predictions and return them.

        Args:
            input (a pandas dataframe): The data on which to do the predictions. There will be
                one prediction per row in the dataframe"""
        try:
            logger.info("Model = {}".format(self.model))
            # TODO replace this with your model.predict
            pred = np.array(['pred_1', 'pred_2', 'pred_3'])
            logger.info(pred)
            return pred
        except Exception as exception:
            # get stack trace as exception
            stack_trace = extract_stack_trace()
            reformatted_log_msg = (
                    'Server Error: {ex}'.format(ex=stack_trace)
            )
            return reformatted_log_msg


def handler(json_input):
    """
    Prediction given the request input
    :param json_input: [dict], request input
    :return: [dict], prediction
    """
    
    model_service = ModelService(model_path=os.path.join(environment.model_dir, "model.joblib"))

    # TODO input transformer
    model_input = model_service.input_transform(input)

    # TODO model prediction
    pred = model_service.predict(model_input)
    logger.info(pred)
    # TODO Add any output processing
    results = model_service.output_transform(
        predictions=pred
    )

    return results



"""
    AWS Auth utilities
"""
import base64
import boto3

from mldock.platform_helpers.utils import strip_scheme

def get_aws_ecr(region: str):
    """
        Get Authentication credentials for AWS ECR using boto3

        args:
            region (str): AWS region of the registry to authenticate
        
        return:
            username (str): username to use in docker client auth
            password (str): password token to use in docker client auth
            registry (str): registry url to use
    """
    # this loads AWS access token and secret from env and returns an ECR client
    ecr_client = boto3.client('ecr', region_name=region)
    # get login token
    token = ecr_client.get_authorization_token()
    username, password = base64.b64decode(token['authorizationData'][0]['authorizationToken']).decode().split(':')
    password = base64.b64decode(token["authorizationData"][0]["authorizationToken"]).decode().split(":")[1]
    registry = token['authorizationData'][0]['proxyEndpoint']
    cloud_repository = strip_scheme(registry)
    # return docker credentials
    return username, password, registry, cloud_repository



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module contains constants representing IANA media types."""
JSON = "application/json"
CSV = "text/csv"
OCTET_STREAM = "application/octet-stream"
ANY = "*/*"
NPY = "application/x-npy"
UTF8_TYPES = [JSON, CSV]



"""
    NUMPY CONTENT ENCODERS

    Handle Encoding of Content from numpy array in to appropriate format for response
"""
# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
# import csv
import io
import json

import numpy as np
from scipy.sparse import issparse
from six import BytesIO, StringIO

from mldock.platform_helpers.mldock.inference import content_types
from mldock.platform_helpers.mldock.inference.recordio import (
    _write_numpy_to_dense_tensor,
    _write_spmatrix_to_sparse_tensor,
)

def array_to_npy(array_like, **kwargs):
    """Convert an array-like object to the NPY format.
    To understand what an array-like object is, please see:
    https://docs.scipy.org/doc/numpy/user/basics.creation.html#converting-python-array-like-objects-to-numpy-arrays
    Args:
        array_like (np.array or Iterable or int or float): Array-like object to be converted to NPY.
    Returns:
        (obj): NPY array.
    """
    buffer = BytesIO()
    np.save(buffer, array_like)
    return buffer.getvalue()

def array_to_csv(array_like, quoted=True, **kwargs):
    """Convert an array like object to CSV.
    To understand what an array-like object is, please see:
    https://docs.scipy.org/doc/numpy/user/basics.creation.html#converting-python-array-like-objects-to-numpy-arrays
    Args:
        array_like (np.array or Iterable or int or float): Array-like object to be converted to CSV.
    Returns:
        (str): Object serialized to CSV.
    """
    data = np.array(array_like)
    if len(data.shape) == 1:
        data = np.reshape(data, (data.shape[0], 1))

    stream = io.StringIO()

    print(data)
    if quoted:
        np.savetxt(stream, data, delimiter=",", fmt='"%s"')
    else:
        np.savetxt(stream, data, delimiter=",", fmt='%s')
    return stream.getvalue()

def array_to_json(array_like):
    """
        Convert an array-like object to JSON.

        To understand what an array-like object is, please see:
        https://docs.scipy.org/doc/numpy/user/basics.creation.html#converting-python-array-like-objects-to-numpy-arrays
        Args:
            array_like (np.array or Iterable or int or float): Array-like object to be
                                                            converted to JSON.
        Returns:
            (str): Object serialized to JSON.
    """

    def default(_array_like):
        if hasattr(_array_like, "tolist"):
            return _array_like.tolist()
        return json.JSONEncoder().default(_array_like)

    return json.dumps(array_like, default=default)

def array_to_recordio_protobuf(array_like, labels=None):
    """Convert an array like object to recordio-protobuf format.
    To understand what an array-like object is, please see:
    https://docs.scipy.org/doc/numpy/user/basics.creation.html#converting-python-array-like-objects-to-numpy-arrays
     Args:
        array_like (np.array or scipy.sparse.csr_matrix): Array-like object to be
                                                          converted to recordio-protobuf.
        labels (np.array or scipy.sparse.csr_matrix): Array-like object representing
                                                      the labels to be encoded.
    Returns:
        buffer: Bytes buffer recordio-protobuf.
    """

    if len(array_like.shape) == 1:
        array_like = array_like.reshape(1, array_like.shape[0])
    assert len(array_like.shape) == 2, "Expecting a 1 or 2 dimensional array"

    buffer = io.BytesIO()

    if issparse(array_like):
        _write_spmatrix_to_sparse_tensor(buffer, array_like, labels)
    else:
        _write_numpy_to_dense_tensor(buffer, array_like, labels)
    buffer.seek(0)
    return buffer.getvalue()



""" Test Utility methods used in platform helpers module"""
import json
import tempfile
from pathlib import Path
from mldock.platform_helpers import utils

class TestUtilsStorage:
    """ Collection of Tests for Utility methods used in platform helpers module"""

    @staticmethod
    def test_get_env_vars_success_on_filter():
        """
            Tests get_env_vars for successfully
            getting all env_vars on filter
        """
        env_vars = {
            "MYTESTVAR1": "2550",
            "_MYTESTVAR2": "a value",
            "MYTESTVAR3_": json.dumps(["a", "list", "example"]),
            "MYTESTVAR4": json.dumps({"key": "value"})
        }
        valid_vars = [{"key": key, 'value': value} for key,value in env_vars.items()]

        with utils.set_env(**env_vars):

            result = utils.get_env_vars(env_vars, ".*MYTESTVAR.*")

        assert result == valid_vars, (
            "Failed. get_env_vars did not produce the expected vars."
        )

    @staticmethod
    def test_delete_file_successful():
        """test delete file successfully removes file"""
        with tempfile.TemporaryDirectory() as tempdir:
            # create file
            file_path = Path(tempdir,'TEST_FILE.txt')
            file_path.touch()

            utils._delete_file(file_path)

            assert (Path(tempdir,'TEST_FILE.txt').is_file() == False), "Fail. Delete file was not successful"

    @staticmethod
    def test_mkdir_successful():
        """test make directory is successfully"""
        with tempfile.TemporaryDirectory() as tempdir:
            # create file
            file_path = Path(tempdir,'TESTPATH')
            
            utils._mkdir(dir_path=file_path)

            assert Path(tempdir,'TESTPATH').is_dir(), "Fail. Make directory was not successful"

    @staticmethod
    def test_list_nested_directories_successful():
        """test listing nested directories successfully"""
        with tempfile.TemporaryDirectory() as tempdir:
            # create file
            file_path = Path(tempdir, 'TESTPATH')
            
            utils._mkdir(dir_path=file_path)

            nested_files = [p.as_posix() for p in utils._iter_nested_dir(tempdir)]

            assert nested_files == [file_path.as_posix()], "Fail. Listing nested directories was not successful"

    @staticmethod
    def test_check_if_cloud_scheme():
        """tests method for checking that string is a s3 url"""
        url = 'gs://bucket/my/path/to/file.txt'

        assert utils._check_if_cloud_scheme(url, scheme='gs'), "Fail. check for gs url scheme was not successful"

    @staticmethod
    def test_strip_scheme():
        """tests method for checking that string is a s3 url"""
        url = 'gs://bucket/my/path/to/file.txt'
        validation = 'bucket/my/path/to/file.txt'

        assert utils.strip_scheme(url) == validation, "Fail. strip scheme from url was not successful"

    @staticmethod
    def test_parse_url_is_successful():
        """test method to parse url to bucket and filepath"""
        url = 'gs://bucket/my/path/to/file.txt'
        bucket, filepath = utils.parse_url(url, scheme='gs')

        assert bucket == 'bucket', "Fail. Parsing url for bucket was not successful"
        assert filepath == 'my/path/to/file.txt', "Fail. Parsing url for filepath was not successful"

    @staticmethod
    def test_zip_and_remove_original_folder_successful():

        with tempfile.TemporaryDirectory() as tempdir:
            # create file
            file_path = Path(tempdir, 'TEST_PATH')
            
            utils._mkdir(dir_path=file_path)
            Path(file_path, 'TEST_FILE.txt').touch()

            output_file = Path(file_path, 'zipped.zip')

            # create and zip folder as .zip
            utils.zip_folder(file_path, output_file, rm_original=True)

            output_file_list = [p.as_posix() for p in utils._iter_nested_dir(file_path)]

            assert output_file.is_file(), "Fail. Zip and remove original was not successful"
            assert len(output_file_list) == 1, "Fail. Zip and remove original was not successful"

    @staticmethod
    def test_zip_to_tarfile_and_rm_original_folder_successful():

        with tempfile.TemporaryDirectory() as tempdir:
            # create file
            file_path = Path(tempdir,'TESTPATH')
            
            utils._mkdir(dir_path=file_path)
            Path(file_path, 'TEST_FILE.txt').touch()

            output_file = Path(file_path, 'zipped.tar.gz')

            # create and zip folder as .tar.gz
            utils.zip_folder_as_tarfile(file_path, output_file, rm_original=True)

            output_file_list = [p.as_posix() for p in utils._iter_nested_dir(file_path)]

            assert output_file.is_file(), "Fail. Zip (tar) and remove original was not successful"
            assert len(output_file_list) == 1, "Fail. Zip (tar) and remove original was not successful"


    @staticmethod
    def test_unzip_and_remove_original_folder_successful():

        with tempfile.TemporaryDirectory() as tempdir:
            # create file
            file_path = Path(tempdir, 'TEST_PATH')
            
            utils._mkdir(dir_path=file_path)
            Path(file_path, 'TEST_FILE.txt').touch()

            output_file = Path(file_path, 'zipped.zip')

            # create and zip folder as .zip
            utils.zip_folder(file_path, output_file, rm_original=True)

            utils.unzip_file(output_file, file_path, rm_zipped=True)

            output_file_list = [p.name for p in utils._iter_nested_dir(file_path)]

            assert output_file_list == ['TEST_FILE.txt'], "Fail. Zip and remove original was not successful"

    @staticmethod
    def test_unzip_file_from_tarfile_and_rm_original_folder_successful():

        with tempfile.TemporaryDirectory() as tempdir:
            # create file
            file_path = Path(tempdir,'TESTPATH')
            
            utils._mkdir(dir_path=file_path)
            Path(file_path, 'TEST_FILE.txt').touch()

            output_file = Path(file_path, 'zipped.tar.gz')

            # create and zip folder as .tar.gz
            utils.zip_folder_as_tarfile(file_path, output_file, rm_original=True)

            utils.unzip_file_from_tarfile(output_file, file_path, rm_zipped=True)

            output_file_list = [p.name for p in utils._iter_nested_dir(file_path)]

            assert output_file_list == ['TEST_FILE.txt'], "Fail. Zip and remove original was not successful"



"""Test registry cli commands"""
from pathlib import Path
import tempfile
from mock import patch
from click.testing import CliRunner

from mldock.__main__ import cli
from mldock.platform_helpers import utils


class TestRegistryCommands:
    """Test Registry cli commands"""
    @staticmethod
    @patch('mldock.command.registry.login_and_authenticate', return_value=(None, {"repository": None}))
    @patch('mldock.command.registry.docker_build')
    @patch('mldock.command.registry.push_image_to_repository')
    def test_push_build_gets_correct_requirements_from_config(
        push_image_mock,
        docker_build_mock,
        docker_auth_mock
    ):
        """test build get's correct requirements from config"""
        runner = CliRunner()

        with tempfile.TemporaryDirectory() as tmp_dir:

            utils._copy_boilerplate_to_dst(
                src='tests/command/fixtures/base_container',
                dst=tmp_dir,
                remove_first=True
            )

            mldock_config = utils._read_json("tests/command/fixtures/base_container/mldock.json")

            result = runner.invoke(
                cli=cli,
                args=['configure', 'local'],
                input='2\n\n'
            )

            _ = runner.invoke(cli=cli, args=['container', 'init', '--dir', tmp_dir, '--no-prompt'])

            _ = runner.invoke(
                cli=cli,
                args=[
                    'registry',
                    'push',
                    '--dir',
                    tmp_dir,
                    '--build',
                    '--provider',
                    'dockerhub',
                    '--region',
                    None
                ]
            )

            _, kwargs = docker_build_mock.call_args
            requirements_path = Path(kwargs['requirements_file_path']).relative_to(tmp_dir).as_posix()

            assert requirements_path == mldock_config["requirements_dir"], (
                "Failed to get correct requirements directory for build"
            )



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module contains utility functions used to generate recordio-protobuf format."""
import struct
import sys

import numpy as np
from scipy.sparse import issparse

from mldock.platform_helpers.mldock.inference.record_pb2 import Record


def _resolve_type(dtype):
    """Return the type string corresponding to the numpy.dtype.
    Args:
        dtype (numpy.dtype or str): numpy.dtype object.
    Returns:
        (str): String corresponding to the dtype.
    """
    if dtype == np.dtype(int):
        return "Int32"
    if dtype == np.dtype(float):
        return "Float64"
    if dtype == np.dtype("float32"):
        return "Float32"
    raise ValueError("Unsupported dtype {} on array".format(dtype))


def _write_feature_tensor(resolved_type, record, vector):
    """Write the feature tensor in the record based on the resolved type.
    Args:
        resolved_type (str): String representing the feature type.
        record (Record object): Record object to write to.
        vector (np.array or csr_matrix): Represents the row (1D Array).
    """
    if resolved_type == "Int32":
        record.features["values"].int32_tensor.values.extend(vector)
    elif resolved_type == "Float64":
        record.features["values"].float64_tensor.values.extend(vector)
    elif resolved_type == "Float32":
        record.features["values"].float32_tensor.values.extend(vector)


def _write_label_tensor(resolved_type, record, scalar):
    """Writes the label to record based on the resolved type.
    Args:
        resolved_type (str): String representing the feature type.
        record (Record object): Record object.
        scalar (int or float32 or float64): Label value.
    """
    if resolved_type == "Int32":
        record.label["values"].int32_tensor.values.extend([scalar])
    elif resolved_type == "Float64":
        record.label["values"].float64_tensor.values.extend([scalar])
    elif resolved_type == "Float32":
        record.label["values"].float32_tensor.values.extend([scalar])


def _write_keys_tensor(resolved_type, record, vector):
    """Write the keys entries in the Record object.
    Args:
        resolved_type (str): Representing the type of key entry.
        record (Record object): Record to which the key will be added.
        vector (array): Array of keys to be added.
    """
    if resolved_type == "Int32":
        record.features["values"].int32_tensor.keys.extend(vector)
    elif resolved_type == "Float64":
        record.features["values"].float64_tensor.keys.extend(vector)
    elif resolved_type == "Float32":
        record.features["values"].float32_tensor.keys.extend(vector)


def _write_shape(resolved_type, record, scalar):
    """Writes the shape entry in the Record.
    Args:
        resolved_type (str): Representing the type of key entry.
        record (Record object): Record to which the key will be added.
        scalar (int or float32 or float64): The shape to added to the record.
    """
    if resolved_type == "Int32":
        record.features["values"].int32_tensor.shape.extend([scalar])
    elif resolved_type == "Float64":
        record.features["values"].float64_tensor.shape.extend([scalar])
    elif resolved_type == "Float32":
        record.features["values"].float32_tensor.shape.extend([scalar])


def _write_numpy_to_dense_tensor(file, array, labels=None):
    """Writes a numpy array to a dense record.
    Args:
        file (file-like object): File-like object where the
                                 records will be written.
        array (numpy array): Numpy array containing the features.
        labels (numpy array): Numpy array containing the labels.
    """

    # Validate shape of array and labels, resolve array and label types
    if not len(array.shape) == 2:
        raise ValueError("Array must be a Matrix")
    if labels is not None:
        if not len(labels.shape) == 1:
            raise ValueError("Labels must be a Vector")
        if labels.shape[0] not in array.shape:
            raise ValueError(
                "Label shape {} not compatible with array shape {}".format(
                    labels.shape, array.shape
                )
            )
        resolved_label_type = _resolve_type(labels.dtype)
    resolved_type = _resolve_type(array.dtype)

    # Write each vector in array into a Record in the file object
    record = Record()
    for index, vector in enumerate(array):
        record.Clear()
        _write_feature_tensor(resolved_type, record, vector)
        if labels is not None:
            _write_label_tensor(resolved_label_type, record, labels[index])
        _write_recordio(file, record.SerializeToString())


def _write_spmatrix_to_sparse_tensor(file, array, labels=None):
    """Writes a scipy sparse matrix to a sparse tensor.
    Args:
        file (file-like object): File-like object where the
                                 records will be written.
        array (array-like): A sparse matrix containing features.
        labels (numpy array): Numpy array containing the labels.
    """

    if not issparse(array):
        raise TypeError("Array must be sparse")

    # Validate shape of array and labels, resolve array and label types
    if not len(array.shape) == 2:
        raise ValueError("Array must be a Matrix")
    if labels is not None:
        if not len(labels.shape) == 1:
            raise ValueError("Labels must be a Vector")
        if labels.shape[0] not in array.shape:
            raise ValueError(
                "Label shape {} not compatible with array shape {}".format(
                    labels.shape, array.shape
                )
            )
        resolved_label_type = _resolve_type(labels.dtype)
    resolved_type = _resolve_type(array.dtype)

    csr_array = array.tocsr()
    n_rows, n_cols = csr_array.shape

    record = Record()
    for row_idx in range(n_rows):
        record.Clear()
        row = csr_array.getrow(row_idx)
        # Write values
        _write_feature_tensor(resolved_type, record, row.data)
        # Write keys
        _write_keys_tensor(resolved_type, record, row.indices.astype(np.uint64))

        # Write labels
        if labels is not None:
            _write_label_tensor(resolved_label_type, record, labels[row_idx])

        # Write shape
        _write_shape(resolved_type, record, n_cols)

        _write_recordio(file, record.SerializeToString())


# MXNet requires recordio records have length in bytes that's a multiple of 4
# This sets up padding bytes to append to the end of the record, for diferent
# amounts of padding required.
padding = {}
for amount in range(4):
    if sys.version_info >= (3,):
        padding[amount] = bytes([0x00 for _ in range(amount)])
    else:
        padding[amount] = bytearray([0x00 for _ in range(amount)])

_kmagic = 0xCED7230A


def _write_recordio(f, data):
    """Wraps the data with RecordIO magic and writes to file-like object.
    Args:
        f (file-like object): The file-like object to which the data point will be written.
        data (numpy array): Data to write.
    """
    length = len(data)
    f.write(struct.pack("I", _kmagic))
    f.write(struct.pack("I", length))
    pad = (((length + 3) >> 2) << 2) - length
    f.write(data)
    f.write(padding[pad])


def _read_recordio(f):
    """Reads a RecordIO and unpacks the body.
    Args:
        f: File like object.
    """
    while True:
        try:
            read_kmagic, = struct.unpack("I", f.read(4))
        except struct.error:
            return
        assert read_kmagic == _kmagic
        len_record, = struct.unpack("I", f.read(4))
        pad = (((len_record + 3) >> 2) << 2) - len_record
        yield f.read(len_record)
        if pad:
            f.read(pad)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import os
import sys
import click

from mldock.__version__ import __version__
from mldock.__init__ import MLDOCK_LOGO
from mldock.command.container import container
from mldock.command.configure import configure
from mldock.command.local import local
from mldock.command.registry import registry
from mldock.command.templates import templates
from mldock.log import configure_logger

click.disable_unicode_literals_warning = True

CLI_VERSION = 'Version: cli {}'.format(__version__)

@click.group()
@click.version_option(message='{}\n{}'.format(MLDOCK_LOGO, CLI_VERSION))
@click.option(u"-v", u"--verbose", count=True, help=u"Turn on debug logging")
@click.pass_context
def cli(ctx, verbose):
    """
    A CLI that helps put machine learning in places that empower ml developers
    """
    mldock_package_path = os.path.dirname(os.path.realpath(__file__))
    logger=configure_logger('mldock', verbose=verbose)
    logger.info(MLDOCK_LOGO)
    logger.info(CLI_VERSION)
    ctx.obj = {'verbose': verbose, 'mldock_package_path': mldock_package_path, 'logo': MLDOCK_LOGO}

def add_commands(cli):

    cli.add_command(configure)
    cli.add_command(container)
    cli.add_command(local)
    cli.add_command(registry)
    cli.add_command(templates)

add_commands(cli)

if __name__ == '__main__':
    cli()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: record.proto

import sys

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="record.proto",
    package="aialgs.data",
    syntax="proto2",
    serialized_pb=_b(
        '\n\x0crecord.proto\x12\x0b\x61ialgs.data"H\n\rFloat32Tensor\x12\x12\n\x06values\x18\x01 \x03(\x02\x42\x02\x10\x01\x12\x10\n\x04keys\x18\x02 \x03(\x04\x42\x02\x10\x01\x12\x11\n\x05shape\x18\x03 \x03(\x04\x42\x02\x10\x01"H\n\rFloat64Tensor\x12\x12\n\x06values\x18\x01 \x03(\x01\x42\x02\x10\x01\x12\x10\n\x04keys\x18\x02 \x03(\x04\x42\x02\x10\x01\x12\x11\n\x05shape\x18\x03 \x03(\x04\x42\x02\x10\x01"F\n\x0bInt32Tensor\x12\x12\n\x06values\x18\x01 \x03(\x05\x42\x02\x10\x01\x12\x10\n\x04keys\x18\x02 \x03(\x04\x42\x02\x10\x01\x12\x11\n\x05shape\x18\x03 \x03(\x04\x42\x02\x10\x01",\n\x05\x42ytes\x12\r\n\x05value\x18\x01 \x03(\x0c\x12\x14\n\x0c\x63ontent_type\x18\x02 \x01(\t"\xd3\x01\n\x05Value\x12\x34\n\x0e\x66loat32_tensor\x18\x02 \x01(\x0b\x32\x1a.aialgs.data.Float32TensorH\x00\x12\x34\n\x0e\x66loat64_tensor\x18\x03 \x01(\x0b\x32\x1a.aialgs.data.Float64TensorH\x00\x12\x30\n\x0cint32_tensor\x18\x07 \x01(\x0b\x32\x18.aialgs.data.Int32TensorH\x00\x12#\n\x05\x62ytes\x18\t \x01(\x0b\x32\x12.aialgs.data.BytesH\x00\x42\x07\n\x05value"\xa9\x02\n\x06Record\x12\x33\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32!.aialgs.data.Record.FeaturesEntry\x12-\n\x05label\x18\x02 \x03(\x0b\x32\x1e.aialgs.data.Record.LabelEntry\x12\x0b\n\x03uid\x18\x03 \x01(\t\x12\x10\n\x08metadata\x18\x04 \x01(\t\x12\x15\n\rconfiguration\x18\x05 \x01(\t\x1a\x43\n\rFeaturesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.aialgs.data.Value:\x02\x38\x01\x1a@\n\nLabelEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.aialgs.data.Value:\x02\x38\x01\x42\x30\n com.amazonaws.aialgorithms.protoB\x0cRecordProtos'
    ),
)


_FLOAT32TENSOR = _descriptor.Descriptor(
    name="Float32Tensor",
    full_name="aialgs.data.Float32Tensor",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="values",
            full_name="aialgs.data.Float32Tensor.values",
            index=0,
            number=1,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b("\020\001")),
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="keys",
            full_name="aialgs.data.Float32Tensor.keys",
            index=1,
            number=2,
            type=4,
            cpp_type=4,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b("\020\001")),
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="shape",
            full_name="aialgs.data.Float32Tensor.shape",
            index=2,
            number=3,
            type=4,
            cpp_type=4,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b("\020\001")),
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[],
    serialized_start=29,
    serialized_end=101,
)


_FLOAT64TENSOR = _descriptor.Descriptor(
    name="Float64Tensor",
    full_name="aialgs.data.Float64Tensor",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="values",
            full_name="aialgs.data.Float64Tensor.values",
            index=0,
            number=1,
            type=1,
            cpp_type=5,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b("\020\001")),
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="keys",
            full_name="aialgs.data.Float64Tensor.keys",
            index=1,
            number=2,
            type=4,
            cpp_type=4,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b("\020\001")),
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="shape",
            full_name="aialgs.data.Float64Tensor.shape",
            index=2,
            number=3,
            type=4,
            cpp_type=4,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b("\020\001")),
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[],
    serialized_start=103,
    serialized_end=175,
)


_INT32TENSOR = _descriptor.Descriptor(
    name="Int32Tensor",
    full_name="aialgs.data.Int32Tensor",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="values",
            full_name="aialgs.data.Int32Tensor.values",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b("\020\001")),
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="keys",
            full_name="aialgs.data.Int32Tensor.keys",
            index=1,
            number=2,
            type=4,
            cpp_type=4,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b("\020\001")),
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="shape",
            full_name="aialgs.data.Int32Tensor.shape",
            index=2,
            number=3,
            type=4,
            cpp_type=4,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b("\020\001")),
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[],
    serialized_start=177,
    serialized_end=247,
)


_BYTES = _descriptor.Descriptor(
    name="Bytes",
    full_name="aialgs.data.Bytes",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="value",
            full_name="aialgs.data.Bytes.value",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="content_type",
            full_name="aialgs.data.Bytes.content_type",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[],
    serialized_start=249,
    serialized_end=293,
)


_VALUE = _descriptor.Descriptor(
    name="Value",
    full_name="aialgs.data.Value",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="float32_tensor",
            full_name="aialgs.data.Value.float32_tensor",
            index=0,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="float64_tensor",
            full_name="aialgs.data.Value.float64_tensor",
            index=1,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="int32_tensor",
            full_name="aialgs.data.Value.int32_tensor",
            index=2,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="bytes",
            full_name="aialgs.data.Value.bytes",
            index=3,
            number=9,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="value",
            full_name="aialgs.data.Value.value",
            index=0,
            containing_type=None,
            fields=[],
        )
    ],
    serialized_start=296,
    serialized_end=507,
)


_RECORD_FEATURESENTRY = _descriptor.Descriptor(
    name="FeaturesEntry",
    full_name="aialgs.data.Record.FeaturesEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="aialgs.data.Record.FeaturesEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="aialgs.data.Record.FeaturesEntry.value",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b("8\001")),
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[],
    serialized_start=674,
    serialized_end=741,
)

_RECORD_LABELENTRY = _descriptor.Descriptor(
    name="LabelEntry",
    full_name="aialgs.data.Record.LabelEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="aialgs.data.Record.LabelEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="aialgs.data.Record.LabelEntry.value",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b("8\001")),
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[],
    serialized_start=743,
    serialized_end=807,
)

_RECORD = _descriptor.Descriptor(
    name="Record",
    full_name="aialgs.data.Record",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="features",
            full_name="aialgs.data.Record.features",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="label",
            full_name="aialgs.data.Record.label",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="uid",
            full_name="aialgs.data.Record.uid",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="metadata",
            full_name="aialgs.data.Record.metadata",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="configuration",
            full_name="aialgs.data.Record.configuration",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[_RECORD_FEATURESENTRY, _RECORD_LABELENTRY],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto2",
    extension_ranges=[],
    oneofs=[],
    serialized_start=510,
    serialized_end=807,
)

_VALUE.fields_by_name["float32_tensor"].message_type = _FLOAT32TENSOR
_VALUE.fields_by_name["float64_tensor"].message_type = _FLOAT64TENSOR
_VALUE.fields_by_name["int32_tensor"].message_type = _INT32TENSOR
_VALUE.fields_by_name["bytes"].message_type = _BYTES
_VALUE.oneofs_by_name["value"].fields.append(_VALUE.fields_by_name["float32_tensor"])
_VALUE.fields_by_name["float32_tensor"].containing_oneof = _VALUE.oneofs_by_name["value"]
_VALUE.oneofs_by_name["value"].fields.append(_VALUE.fields_by_name["float64_tensor"])
_VALUE.fields_by_name["float64_tensor"].containing_oneof = _VALUE.oneofs_by_name["value"]
_VALUE.oneofs_by_name["value"].fields.append(_VALUE.fields_by_name["int32_tensor"])
_VALUE.fields_by_name["int32_tensor"].containing_oneof = _VALUE.oneofs_by_name["value"]
_VALUE.oneofs_by_name["value"].fields.append(_VALUE.fields_by_name["bytes"])
_VALUE.fields_by_name["bytes"].containing_oneof = _VALUE.oneofs_by_name["value"]
_RECORD_FEATURESENTRY.fields_by_name["value"].message_type = _VALUE
_RECORD_FEATURESENTRY.containing_type = _RECORD
_RECORD_LABELENTRY.fields_by_name["value"].message_type = _VALUE
_RECORD_LABELENTRY.containing_type = _RECORD
_RECORD.fields_by_name["features"].message_type = _RECORD_FEATURESENTRY
_RECORD.fields_by_name["label"].message_type = _RECORD_LABELENTRY
DESCRIPTOR.message_types_by_name["Float32Tensor"] = _FLOAT32TENSOR
DESCRIPTOR.message_types_by_name["Float64Tensor"] = _FLOAT64TENSOR
DESCRIPTOR.message_types_by_name["Int32Tensor"] = _INT32TENSOR
DESCRIPTOR.message_types_by_name["Bytes"] = _BYTES
DESCRIPTOR.message_types_by_name["Value"] = _VALUE
DESCRIPTOR.message_types_by_name["Record"] = _RECORD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Float32Tensor = _reflection.GeneratedProtocolMessageType(
    "Float32Tensor",
    (_message.Message,),
    dict(
        DESCRIPTOR=_FLOAT32TENSOR,
        __module__="record_pb2"
        # @@protoc_insertion_point(class_scope:aialgs.data.Float32Tensor)
    ),
)
_sym_db.RegisterMessage(Float32Tensor)

Float64Tensor = _reflection.GeneratedProtocolMessageType(
    "Float64Tensor",
    (_message.Message,),
    dict(
        DESCRIPTOR=_FLOAT64TENSOR,
        __module__="record_pb2"
        # @@protoc_insertion_point(class_scope:aialgs.data.Float64Tensor)
    ),
)
_sym_db.RegisterMessage(Float64Tensor)

Int32Tensor = _reflection.GeneratedProtocolMessageType(
    "Int32Tensor",
    (_message.Message,),
    dict(
        DESCRIPTOR=_INT32TENSOR,
        __module__="record_pb2"
        # @@protoc_insertion_point(class_scope:aialgs.data.Int32Tensor)
    ),
)
_sym_db.RegisterMessage(Int32Tensor)

Bytes = _reflection.GeneratedProtocolMessageType(
    "Bytes",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BYTES,
        __module__="record_pb2"
        # @@protoc_insertion_point(class_scope:aialgs.data.Bytes)
    ),
)
_sym_db.RegisterMessage(Bytes)

Value = _reflection.GeneratedProtocolMessageType(
    "Value",
    (_message.Message,),
    dict(
        DESCRIPTOR=_VALUE,
        __module__="record_pb2"
        # @@protoc_insertion_point(class_scope:aialgs.data.Value)
    ),
)
_sym_db.RegisterMessage(Value)

Record = _reflection.GeneratedProtocolMessageType(
    "Record",
    (_message.Message,),
    dict(
        FeaturesEntry=_reflection.GeneratedProtocolMessageType(
            "FeaturesEntry",
            (_message.Message,),
            dict(
                DESCRIPTOR=_RECORD_FEATURESENTRY,
                __module__="record_pb2"
                # @@protoc_insertion_point(class_scope:aialgs.data.Record.FeaturesEntry)
            ),
        ),
        LabelEntry=_reflection.GeneratedProtocolMessageType(
            "LabelEntry",
            (_message.Message,),
            dict(
                DESCRIPTOR=_RECORD_LABELENTRY,
                __module__="record_pb2"
                # @@protoc_insertion_point(class_scope:aialgs.data.Record.LabelEntry)
            ),
        ),
        DESCRIPTOR=_RECORD,
        __module__="record_pb2"
        # @@protoc_insertion_point(class_scope:aialgs.data.Record)
    ),
)
_sym_db.RegisterMessage(Record)
_sym_db.RegisterMessage(Record.FeaturesEntry)
_sym_db.RegisterMessage(Record.LabelEntry)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(
    descriptor_pb2.FileOptions(), _b("\n com.amazonaws.aialgorithms.protoB\014RecordProtos")
)
_FLOAT32TENSOR.fields_by_name["values"].has_options = True
_FLOAT32TENSOR.fields_by_name["values"]._options = _descriptor._ParseOptions(
    descriptor_pb2.FieldOptions(), _b("\020\001")
)
_FLOAT32TENSOR.fields_by_name["keys"].has_options = True
_FLOAT32TENSOR.fields_by_name["keys"]._options = _descriptor._ParseOptions(
    descriptor_pb2.FieldOptions(), _b("\020\001")
)
_FLOAT32TENSOR.fields_by_name["shape"].has_options = True
_FLOAT32TENSOR.fields_by_name["shape"]._options = _descriptor._ParseOptions(
    descriptor_pb2.FieldOptions(), _b("\020\001")
)
_FLOAT64TENSOR.fields_by_name["values"].has_options = True
_FLOAT64TENSOR.fields_by_name["values"]._options = _descriptor._ParseOptions(
    descriptor_pb2.FieldOptions(), _b("\020\001")
)
_FLOAT64TENSOR.fields_by_name["keys"].has_options = True
_FLOAT64TENSOR.fields_by_name["keys"]._options = _descriptor._ParseOptions(
    descriptor_pb2.FieldOptions(), _b("\020\001")
)
_FLOAT64TENSOR.fields_by_name["shape"].has_options = True
_FLOAT64TENSOR.fields_by_name["shape"]._options = _descriptor._ParseOptions(
    descriptor_pb2.FieldOptions(), _b("\020\001")
)
_INT32TENSOR.fields_by_name["values"].has_options = True
_INT32TENSOR.fields_by_name["values"]._options = _descriptor._ParseOptions(
    descriptor_pb2.FieldOptions(), _b("\020\001")
)
_INT32TENSOR.fields_by_name["keys"].has_options = True
_INT32TENSOR.fields_by_name["keys"]._options = _descriptor._ParseOptions(
    descriptor_pb2.FieldOptions(), _b("\020\001")
)
_INT32TENSOR.fields_by_name["shape"].has_options = True
_INT32TENSOR.fields_by_name["shape"]._options = _descriptor._ParseOptions(
    descriptor_pb2.FieldOptions(), _b("\020\001")
)
_RECORD_FEATURESENTRY.has_options = True
_RECORD_FEATURESENTRY._options = _descriptor._ParseOptions(
    descriptor_pb2.MessageOptions(), _b("8\001")
)
_RECORD_LABELENTRY.has_options = True
_RECORD_LABELENTRY._options = _descriptor._ParseOptions(
    descriptor_pb2.MessageOptions(), _b("8\001")
)
# @@protoc_insertion_point(module_scope)



from pathlib import Path
from dataclasses import dataclass

@dataclass
class GSBucketObject:
    """Class for mocking the google storage bucket object blob."""
    name: str
    bucket: str

    @staticmethod
    def download_to_filename(local_path):
        print("download_to_filename: {}".format(local_path))

    @staticmethod
    def upload_from_filename(local_path):
        print("upload_to_filename: {}".format(local_path))

@dataclass
class GSBucket:
    """Class for mocking the google storage bucket object."""
    name: str
    objects: list

    def list_blobs(self, prefix) -> list:
        print("listing these blobs")
        return [
            GSBucketObject(bucket=self, name=Path(prefix, item).as_posix())
            for item in self.objects
        ]

    def blob(self, prefix):
        print("running this blob")
        return GSBucketObject(bucket=self, name=prefix)





"""Registry API methods"""
import logging
import click
import docker

logger=logging.getLogger('mldock')

def compute_progress(current, max_value=10):

    if current > max_value:
        multiple = current // max_value
        current = current - max_value * multiple 
    
    return current

def get_layer_state(stream, layer_id, state={}, progress='', fill_char='=', increment=2, max_value=40, status_tag='pushing'):
    """
        Computes a status update to the state of layer during a push.
        Additionally, adds a progress bar and flexible fill character.
    """
    try:

        if status_tag.lower() == 'pushing':
            progress = ''
            if stream.lower() == status_tag.lower():
                current_layer = state[layer_id]
                current_progress = current_layer['progress'].replace("[",'').replace("]",'').replace(">",'')
                current_progress = len(current_progress)
                progress = compute_progress(current_progress, max_value=max_value) + increment
                progress = fill_char * progress
                progress = "[{}>]".format(progress)

        state.update(
            {layer_id: {'message':stream, 'progress': progress}}
        )
    except KeyError as exception:
        pass

    return state

def stateful_log_emitter(line: dict, status_tag: str, states={}):
    error = line.get('error', None)
    errorDetail = line.get('errorDetail', None)
    if error is not None:
        logger.error('{}\n{}'.format(error, errorDetail))
    
    stream = line.get('status','')
    layer_id = line.get('id','')
    progress = line.get('progress','')

    if (layer_id == ''):
        return states
    else:
        # perform status update, using layer_id as key
        states = get_layer_state(stream, layer_id, states, progress=progress, status_tag=status_tag)

        # clear terminal to allow for friendlier update
        click.clear()

        # emit current state to stdout
        {logger.info("{}: {} {}".format(k,v['message'], v['progress'])) for k, v in states.items()}

    return states

def push_image_to_repository(
    image_repository: str,
    auth_config: dict,
    tag='latest'
):
    """
        Push image to repository in registry. Using auth_config this will Authenticate client.
    """
    client = docker.from_env()

    push_response = client.images.push(
        repository=image_repository,
        tag=tag,
        stream=True,
        decode=True,
        auth_config=auth_config
    )

    states = {}
    for line in push_response:
        states = stateful_log_emitter(line=line, states=states, status_tag='Pushing')
    return states

def pull_image_from_repository(
    image_repository: str,
    auth_config: dict,
    tag='latest'
):
    """
        Pull image from repository in registry. Using auth_config this will Authenticate client.
    """
    client = docker.from_env()

    pull_response = client.api.pull(
        repository=image_repository,
        tag=tag,
        stream=True,
        decode=True,
        auth_config=auth_config
    )

    states = {}
    for line in pull_response:
        states = stateful_log_emitter(line=line, states=states, status_tag='Downloading')
    return states



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
#
# Disclaimer: This code can be found here: https://github.com/aws/sagemaker-training-toolkit/blob/master/test/unit/test_encoder.py
#
import io
import itertools
import json

from mock import Mock, patch
# import numpy as np
import pandas as pd
import pytest
from scipy import sparse
from six import BytesIO

from mldock.platform_helpers.mldock.inference import content_types
from mldock.platform_helpers.mldock.inference.content_decoders import pandas as pandas_decoders


class TestPandasDecoders:
    """Tests the pandas decoder methods"""

    @staticmethod
    @pytest.mark.parametrize(
        "target, expected",
        [
            ({"col":[42, 6, 9]}, pd.DataFrame([42, 6, 9], columns=["col"])),
            ({"col":[42.0, 6.0, 9.0]}, pd.DataFrame([42.0, 6.0, 9.0], columns=["col"])),
            ({"col":["42", "6", "9"]}, pd.DataFrame(["42", "6", "9"], columns=["col"])),
            ({"col":[u"42", u"6", u"9"]}, pd.DataFrame([u"42", u"6", u"9"], columns=["col"])),
        ],
    )
    def test_json_list_to_pandas(target, expected):
        """test json data is correctly decoded to pandas dataframe"""
        actual = pandas_decoders.json_list_to_pandas(json.dumps(target))
        pd.testing.assert_frame_equal(actual, expected)

    @staticmethod
    @pytest.mark.parametrize(
        "target, expected",
        [
            (b"col\n42\n6\n9\n", pd.DataFrame([42, 6, 9], columns=['col'])),
            (b"col\n42.0\n6.0\n9.0\n", pd.DataFrame([42.0, 6.0, 9.0], columns=['col'])),
            (b"col\n42\n6\n9\n", pd.DataFrame([42, 6, 9], columns=['col'])),
            (b'col\n"False,"\n"True."\n"False,"\n', pd.DataFrame(["False,", "True.", "False,"], columns=['col'])),
            (b'col\naaa\n"b""bb"\nccc\n', pd.DataFrame(["aaa", 'b"bb', "ccc"], columns=['col'])),
            (b'col\n"a\nb"\nc\n', pd.DataFrame(["a\nb", "c"], columns=['col']))
        ],
    )
    def test_csv_to_pandas(target, expected):
        """test csv data is correctly decoded to pandas dataframe"""
        actual = pandas_decoders.csv_to_pandas(target)
        pd.testing.assert_frame_equal(actual, expected)



import os
import sys
import logging
import click

from mldock.platform_helpers.docker.auth import login_and_authenticate
from mldock.api.local import \
    docker_build
from mldock.api.registry import \
    push_image_to_repository, pull_image_from_repository
from mldock.config_managers.container import \
    MLDockConfigManager
from mldock.terminal import ProgressLogger, pretty_build_logs

click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')
MLDOCK_CONFIG_NAME = 'mldock.json'

def reset_terminal():
    # os.system("clear")
    click.clear()

@click.group()
def registry():
    """
    Commands to interact with docker image registries.
    """
    pass

@click.command()
@click.option(
    '--dir',
    help='Set the working directory for your mldock container.',
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None
    )
)
@click.option('--no-cache', help='builds container from scratch', is_flag=True)
@click.option(
    '--build',
    help='Set the working directory for your mldock container.',
    is_flag=True
)
@click.option(
    '--provider',
    help='Set the cloud provider',
    required=True,
    type=click.Choice(
        ['ecr', 'gcr', 'dockerhub'],
        case_sensitive=False
    )
)
@click.option(
    '--region',
    help='Set the registry region',
    default=None,
    type=str
)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--stage', help='environment to stage.')
@click.pass_obj
def push(obj, dir, region, build, provider, no_cache, stage, tag):
    """
    Command to push docker container image to Image Registry
    """
    reset_terminal()
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(dir, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_dir name
    mldock_config = mldock_manager.get_config()
    image_name = mldock_config.get("image_name", None)
    container_dir = mldock_config.get("container_dir", None)
    module_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
    )
    dockerfile_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
        container_dir
    )
    requirements_file_path = os.path.join(
        dir,
        mldock_config.get("requirements_dir", "requirements.txt")
    )

    # retrieve stages
    with ProgressLogger(
        group='Stages',
        text='Retrieving Stages',
        spinner='dots',
        on_success='Stages Retrieved'
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]['tag']

        if tag is None:
            raise Exception("tag is not valid. Either choose a stage or set a tag manually")

    # athenticate and login docker client
    with ProgressLogger(
        group='Authentication',
        text='Authenticating with {}'.format(provider),
        spinner='dots'
    ) as spinner:
        client, metadata = login_and_authenticate(provider=provider, region=region)
        image_repository = "{}/{}".format(metadata['repository'], image_name)


    if build:
        # build image for cloud repository
        with ProgressLogger(
            group='Build',
            text='Building',
            spinner='dots'
        ) as spinner:
            logs = docker_build(
                image_name=image_repository,
                dockerfile_path=dockerfile_path,
                module_path=module_path,
                target_dir_name=mldock_config.get("mldock_module_dir", "src"),
                requirements_file_path=requirements_file_path,
                no_cache=no_cache,
                docker_tag=tag
            )
            for line in logs:

                spinner.info(pretty_build_logs(line=line))

                spinner.start()

    # Push image to cloud repository
    with ProgressLogger(group='Push', text='Pushing to {}'.format(image_repository), spinner='dots') as spinner:
        states = push_image_to_repository(
            image_repository=image_repository,
            auth_config = {'username': metadata['username'], 'password': metadata['password']},
            tag=tag
        )

        for layer_id, metadata in states.items():

            single_line = "{} {}".format(layer_id, metadata["message"])

            spinner.info(single_line)

        reset_terminal()
        logger.info(obj["logo"])
        spinner.start()

@click.command()
@click.option(
    '--dir',
    help='Set the working directory for your mldock container.',
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None
    )
)
@click.option('--no-cache', help='builds container from scratch', is_flag=True)
@click.option(
    '--provider',
    help='Set the cloud provider',
    required=True,
    type=click.Choice(['ecr', 'gcr', 'dockerhub'],
    case_sensitive=False
    )
)
@click.option(
    '--region',
    help='Set the registry region',
    default=None,
    type=str
)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--stage', help='environment to stage.')
@click.pass_obj
def pull(obj, dir, region, provider, no_cache, stage, tag):
    """
    Command to pull docker container image from Image Registry
    """
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(dir, MLDOCK_CONFIG_NAME)
    )

    # get mldock_module_dir name
    mldock_config = mldock_manager.get_config()
    image_name = mldock_config.get("image_name", None)
    container_dir = mldock_config.get("container_dir", None)
    module_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
    )
    dockerfile_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
        container_dir
    )

    # retrieve stages
    with ProgressLogger(
        group='Stages',
        text='Retrieving Stages',
        spinner='dots',
        on_success='Stages Retrieved'
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]['tag']

        if tag is None:
            raise Exception("tag is not valid. Either choose a stage or set a tag manually")

    # login and authenticate
    with ProgressLogger(
        group='Authentication',
        text='Authenticating with {}'.format(provider),
        spinner='dots'
    ) as spinner:
        client, metadata = login_and_authenticate(provider=provider, region=region)
        image_repository = "{}/{}".format(metadata['repository'], image_name)

    # Push image to cloud repository
    with ProgressLogger(
        group='Pull',
        text='Pulling from {}'.format(image_repository),
        spinner='dots'
    ) as spinner:
        states = pull_image_from_repository(
            image_repository=image_repository,
            auth_config = {'username': metadata['username'], 'password': metadata['password']},
            tag=tag
        )

        for layer_id, metadata in states.items():
            single_line = "{} {}".format(layer_id, metadata["message"])

            spinner.info(single_line)

        reset_terminal()
        logger.info(obj["logo"])
        spinner.start()

registry.add_command(push)
registry.add_command(pull)



"""
    TRAINER SCRIPT - contains custom code for training your model
"""
import os
import sys
import json
import argparse
import pickle

import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import cross_val_score
from sklearn.svm import SVC

from src.container.assets import environment, logger

def str2bool(arg_value: str) -> bool:
    if isinstance(arg_value, bool):
        return arg_value
    if arg_value.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif arg_value.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def str2dict(arg_value: str) -> dict:
    d_argvalue = json.loads(arg_value)
    if isinstance(d_argvalue, dict):
        return d_argvalue
    else:
        raise argparse.ArgumentTypeError('Json Dictionary expected.')

def run_training():

    PARSER= argparse.ArgumentParser()
    ## User set params which link to Sagemaker params
    PARSER.add_argument('--iris', type=str,
                        default=os.path.join(environment.input_data_dir, 'iris'),
                        help='Expects a .csv file, representing the lookup.')
    PARSER.add_argument('--test-size', default=environment.hyperparameters.get('test-size', 0.2),
                        type=int,
                        help='Number of factorization factors to compute during model fit.')
    PARSER.add_argument('--regularization', default=environment.hyperparameters.get('regularization', 0.3),
                        type=int,
                        help='Coefficient of regularization to apply during model fit.')
    PARSER.add_argument('--iterations', default=environment.hyperparameters.get('iterations', 10),
                        type=int,
                        help='Number iterations of the dataset should be run during model fit.')
    PARSER.add_argument('--run_eval', default=environment.hyperparameters.get('run_eval', True),
                        type=str2bool,
                        help='Whether to run metrics evaluation prior to training model on full data.')
    PARSER.add_argument('--refit', default=environment.hyperparameters.get('refit', True),
                        type=str2bool,
                        help='Whether or not to save model.')

    # Get args
    ARGS, _ = PARSER.parse_known_args()

    logger.info("Experiment id = {}".format(environment.environment_variables('MLDOCK_EXPERIMENT_ID', default=None)))

    data = pd.read_csv(os.path.join(ARGS.iris, 'train.csv'))
    X, y = data.drop(['target'], axis=1), data['target']
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=ARGS.test_size, random_state=123, shuffle=True)

    transformer = StandardScaler()
    transformer.fit(X_train)
    X_train_std = transformer.transform(X_train)
    X_test_std = transformer.transform(X_test)

    if ARGS.run_eval:
        logger.info('Running Evaluation')

        model = SVC(kernel='rbf', random_state=0, gamma=.10, C=1.0)
        model.fit(X_train_std, y_train)

        print('Model: Accuracy (train) {:.2f}'.format(model.score(X_train_std, y_train)))
        print('Model: Accuracy (test) {:.2f}'.format(model.score(X_test_std, y_test)))

    if ARGS.refit:
        logger.info('Refitting Final Model')

        transformer = StandardScaler()
        transformer.fit(X)
        X_std = transformer.transform(X)
        model = SVC(kernel='rbf', random_state=0, gamma=.10, C=1.0)
        model.fit(X_std, y)

        logger.info("Model saved in {}. This is written back to remote at job completion.".format(environment.model_dir))
        transformer_path = os.path.join(environment.model_dir, 'iris/transformer.pkl')

        with open(transformer_path, 'wb') as file_:
            pickle.dump(transformer, file_)

        model_path = os.path.join(environment.model_dir, 'iris/model.pkl')

        with open(model_path, 'wb') as file_:
            pickle.dump(model, file_)
    else:
        logger.debug("Skipping model save step. ")

if __name__ == '__main__':

    try:
        run_training()
    except Exception as exception:
        logger.error(exception)



#!/usr/bin/env python

# This file implements the scoring service shell. You don't necessarily need to modify it for
# various algorithms. It starts nginx and gunicorn with the correct configurations and then simply
# waits until gunicorn exits.
#
# The flask server is specified to be the app object in wsgi.py
#
# We set the following parameters:
#
# Parameter                Environment Variable              Default Value
# ---------                --------------------              -------------
# number of workers        MLDOCK_SERVER_WORKERS              the number of CPU cores
# timeout                  MLDOCK_SERVER_TIMEOUT              60 seconds

from __future__ import print_function
import sys;sys.path.insert(1, ".")  # Do not remove this
import multiprocessing
import os
import signal
import sys
from future.moves import subprocess

from src.container.assets import ServingContainer, environment, logger

_DIR_TO_FILE = os.path.dirname(os.path.abspath(__file__))

cpu_count = multiprocessing.cpu_count()

model_server_timeout = environment.environment_variables.int('MLDOCK_SERVER_TIMEOUT', default=60)
model_server_workers = environment.environment_variables.int('MLDOCK_SERVER_WORKERS', default=cpu_count)

def sigterm_handler(nginx_pid, gunicorn_pid):
    try:
        os.kill(nginx_pid, signal.SIGQUIT)
    except OSError:
        pass
    try:
        os.kill(gunicorn_pid, signal.SIGTERM)
    except OSError:
        pass

    sys.exit(0)


def start_server():
    logger.info('Starting the inference server with {} workers.'.format(model_server_workers))

    # link the log streams to stdout/err so they will be logged to the container logs
    subprocess.check_call(['ln', '-sf', '/dev/stdout', '/var/log/nginx/access.log'])
    subprocess.check_call(['ln', '-sf', '/dev/stderr', '/var/log/nginx/error.log'])

    nginx = subprocess.Popen(['nginx', '-c', os.path.join(_DIR_TO_FILE, 'nginx.conf')])
    gunicorn = subprocess.Popen(['gunicorn',
                                 '--timeout', str(model_server_timeout),
                                 '-k', 'gevent',
                                 '-b', 'unix:/tmp/gunicorn.sock',
                                 '-w', str(model_server_workers),
                                 'src.container.prediction.wsgi:app'
                                 ])

    signal.signal(signal.SIGTERM, lambda a, b: sigterm_handler(nginx.pid, gunicorn.pid))

    # If either subprocess exits, so do we.
    pids = set([nginx.pid, gunicorn.pid])
    while True:
        pid, _ = os.wait()
        if pid in pids:
            logger.info("Nginx or gunicorn exited. Exiting server.")
            break

    sigterm_handler(nginx.pid, gunicorn.pid)
    logger.info('Inference server exiting')

# The main routine just invokes the start function.

if __name__ == '__main__':
    serving = ServingContainer(environment=environment)
    try:
        serving.startup()

        start_server()
    except Exception as exception:
        logger.error(exception)
    finally:
        serving.cleanup()



from abc import ABCMeta, abstractmethod


class ModelService(object):
    """
    ModelService wraps up all preprocessing, inference and postprocessing
    functions used by model service. It is defined in a flexible manner to
    be easily extended to support different frameworks.
    """
    __metaclass__ = ABCMeta

    model = None

    def __init__(self, model_path: str):
        """
            Internal initialize ModelService.
            
            args:
                model_path (str): path to model
            return:
                model object
        """
        self.model = self.load_model(model_path)

    @abstractmethod
    def load_model(self, model_path: str):
        """
            (Abstract) load model
        
            args:
                model_path(str): path to model
        """
        # pylint: disable=unnecessary-pass
        pass

    @abstractmethod
    def predict(self, input):
        """
            (Abstract) Predict method

            args:
                input : iterable of records model takes at prediction time.
            return:
                iterable of outputs from model
        """
        # pylint: disable=unnecessary-pass
        pass

    @abstractmethod
    def input_transform(input, **kwargs):
        """
            (Abstract) Output Transform

            args:
                input: raw iterable from request
            return:
                input transformed data
        """
        # pylint: disable=unnecessary-pass
        pass

    @abstractmethod
    def output_transform(predictions, **kwargs):
        """
            (Abstract) Output Transform

            args:
                predictions: predictions returned from model.predict
            return:
                output transformed predictions
        """
        # pylint: disable=unnecessary-pass
        pass



import json

def _format_key_as_mldock_env_var(key, prefix=None):
    """
        Formats key as mldock env variable.
        Replace '\s' and '-', append mldock prefix
        and lastly transforms to uppercase
    """
    if prefix is not None:
        key = "{}_{}".format(prefix, key)
    key = key.replace(" ", "_").replace("-", "_")
    key = key.upper()
    return key

def _format_dictionary_as_env_vars(obj: dict, group=None):
    """
        format key and values as appropriate environment variables

        namely:
            - upper case
            - replace spaces with "_"
            - replace dashes with "_"
            - add owner prefix
    """
    new_keys = {}
    for _key,_value in obj.items():

        _key = _format_key_as_mldock_env_var(_key, prefix=group)

        if isinstance(_value, dict):
            _value = json.dumps(_value)

        new_keys.update(
            {
                _key: _value
            }
        )
    return new_keys

def collect_mldock_environment_variables(**env_vars):
    """Collect and format mldock environment variables"""
    return _format_dictionary_as_env_vars(
        env_vars,
        group='mldock'
    )

def _make_sagemaker_input_data_configs(data_channels: list):
    """restructures a list of data names as sagemaker input data configs"""
    return {name: {} for name in data_channels}

def _extract_data_channels_from_mldock(mldock_data: list):

    data_channels = [
        data_config['channel'] for data_config in mldock_data
    ]

    return _make_sagemaker_input_data_configs(data_channels=data_channels)
